from __future__ import annotations

import sys
from dataclasses import dataclass

from pyqtorch.utils import StrEnum


class DigitalNoiseType(StrEnum):
    BITFLIP = "BitFlip"
    PHASEFLIP = "PhaseFlip"
    DEPOLARIZING = "Depolarizing"
    PAULI_CHANNEL = "PauliChannel"
    AMPLITUDE_DAMPING = "AmplitudeDamping"
    PHASE_DAMPING = "PhaseDamping"
    GENERALIZED_AMPLITUDE_DAMPING = "GeneralizedAmplitudeDamping"


@dataclass
class DigitalNoiseInstance:
    type: DigitalNoiseType
    error_probability: tuple[float, ...] | float | None
    target: int | None


class DigitalNoiseProtocol:
    """
    Define a protocol of noisy quantum channels.

    Args:
        protocol: single DigitalNoiseType instance or list of DigitalNoiseType instances, or
            list of (DigitalNoiseType, options) tuple. When passing list of tuples, for
            each DigitalNoiseType the options should be a dict containing the "error_probability",
            and optionally a "target". If no "target" is present, the noise instance
            will be applied to the same target of the gate it is used on.
        error_probability: probability of error when passing a single DigitalNoiseType
            or a list of NoiseTypes. Note that all noise types require a single float for the
            error_probability, while the PauliChannel and GeneralizedAmplitudeDamping require
            a tuple of error probabilities.

    Examples:
    ```
        from pyqtorch.noise import DigitalNoiseProtocol, DigitalNoiseType

        # Single noise instance
        protocol = DigitalNoiseProtocol(DigitalNoiseType.BITFLIP, error_probability = 0.5)

        # Equivalent to using the respective class method
        protocol = DigitalNoiseProtocol.bitflip(error_probability = 0.5)

        # Multiples noise instances with same probability
        protocol = DigitalNoiseProtocol(
                        [DigitalNoiseType.BITFLIP, DigitalNoiseType.PHASEFLIP],
                        error_probability = 0.5
                    )

        # Multiples noise instances with different options
        prot0 = DigitalNoiseProtocol.bitflip(0.5)
        prot1 = DigitalNoiseProtocol.pauli_channel((0.1, 0.2, 0.7))
        protocol = DigitalNoiseProtocol([prot0, prot1])

    ```
    """

    def __init__(
        self,
        protocol: (
            DigitalNoiseType | list[DigitalNoiseType] | list[DigitalNoiseProtocol]
        ),
        error_probability: tuple[float, ...] | float | None = None,
        target: int | None = None,
    ) -> None:

        self._error_probability = error_probability
        self._target = target

        is_protocol_list = isinstance(protocol, list)

        if isinstance(protocol, DigitalNoiseType):
            self.noise_instances = [
                DigitalNoiseInstance(protocol, error_probability, target)
            ]
        elif is_protocol_list and isinstance(protocol[0], DigitalNoiseType):
            self.noise_instances = [
                DigitalNoiseInstance(p, error_probability, target) for p in protocol  # type: ignore [arg-type]
            ]
        elif is_protocol_list and isinstance(protocol[0], DigitalNoiseProtocol):
            self.noise_instances = []
            for p in protocol:
                self.noise_instances.extend(p.noise_instances)  # type: ignore [union-attr]
        else:
            raise TypeError(f"Incorrect protocol type: {type(protocol)}.")

        for noise in self.noise_instances:
            if noise.error_probability is None:
                raise ValueError(
                    f"No error_probability passed to the protocol {noise.type}."
                )

        self.len = len(self.noise_instances)

    @property
    def error_probability(self):
        if self.len == 1:
            return self._error_probability
        else:
            return [noise.error_probability for noise in self.noise_instances]

    @property
    def target(self):
        if self.len == 1:
            return self._target
        else:
            return [noise.target for noise in self.noise_instances]

    @property
    def gates(self) -> list:
        gate_list = []
        for noise in self.noise_instances:
            try:
                gate_class = getattr(
                    sys.modules["pyqtorch.noise.digital_gates"], str(noise.type)
                )
                gate_list.append((gate_class, noise))
            except AttributeError:
                raise ValueError(
                    f"The protocol {str(noise.type)} has not been implemented in PyQTorch yet."
                )
        return gate_list

    @classmethod
    def bitflip(cls, *args, **kwargs) -> DigitalNoiseProtocol:
        return cls(DigitalNoiseType.BITFLIP, *args, **kwargs)

    @classmethod
    def phaseflip(cls, *args, **kwargs) -> DigitalNoiseProtocol:
        return cls(DigitalNoiseType.PHASEFLIP, *args, **kwargs)

    @classmethod
    def depolarizing(cls, *args, **kwargs) -> DigitalNoiseProtocol:
        return cls(DigitalNoiseType.DEPOLARIZING, *args, **kwargs)

    @classmethod
    def pauli_channel(cls, *args, **kwargs) -> DigitalNoiseProtocol:
        return cls(DigitalNoiseType.PAULI_CHANNEL, *args, **kwargs)

    @classmethod
    def amplitude_damping(cls, *args, **kwargs) -> DigitalNoiseProtocol:
        return cls(DigitalNoiseType.AMPLITUDE_DAMPING, *args, **kwargs)

    @classmethod
    def phase_damping(cls, *args, **kwargs) -> DigitalNoiseProtocol:
        return cls(DigitalNoiseType.PHASE_DAMPING, *args, **kwargs)

    @classmethod
    def generalized_amplitude_damping(cls, *args, **kwargs) -> DigitalNoiseProtocol:
        return cls(DigitalNoiseType.GENERALIZED_AMPLITUDE_DAMPING, *args, **kwargs)

    def __repr__(self) -> str:
        if self.len == 1:
            noise = self.noise_instances[0]
            if noise.target is not None:
                return f"{str(noise.type)}(prob: {noise.error_probability}, target: {noise.target})"
            else:
                return f"{str(noise.type)}(prob: {noise.error_probability})"
        else:
            return f"DigitalNoiseProtocol(length = {self.len})"


def _repr_noise(noise: DigitalNoiseProtocol | None = None) -> str:
    """Returns the string for noise representation in gates."""
    noise_info = ""
    if noise is None:
        return noise_info
    elif isinstance(noise, DigitalNoiseProtocol):
        noise_info = str(noise)
    return f", noise: {noise_info}"



from __future__ import annotations

from .parametric import ControlledParametric, ControlledRotationGate, Parametric
from .parametric_gates import (
    CPHASE,
    CRX,
    CRY,
    CRZ,
    OPS_DIAGONAL_PARAM,
    OPS_PARAM,
    OPS_PARAM_1Q,
    OPS_PARAM_2Q,
    PHASE,
    RX,
    RY,
    RZ,
    U,
)
from .primitive import ControlledPrimitive, Primitive
from .primitive_gates import (
    CNOT,
    CSWAP,
    CY,
    CZ,
    OPS_1Q,
    OPS_2Q,
    OPS_3Q,
    OPS_DIAGONAL,
    OPS_DIGITAL,
    OPS_PAULI,
    SWAP,
    H,
    I,
    N,
    Projector,
    S,
    SDagger,
    T,
    Toffoli,
    X,
    Y,
    Z,
)



from __future__ import annotations

from collections import OrderedDict
from typing import Callable

import pytest
import torch
from helpers import calc_mat_vec_wavefunction, random_pauli_hamiltonian

import pyqtorch as pyq
from pyqtorch import RX, Add, ConcretizedCallable, HamiltonianEvolution, Scale, X
from pyqtorch.composite import Sequence
from pyqtorch.hamiltonians import GeneratorType
from pyqtorch.matrices import (
    DEFAULT_MATRIX_DTYPE,
    DEFAULT_REAL_DTYPE,
    IMAT,
    XMAT,
    ZMAT,
)
from pyqtorch.noise import AnalogDepolarizing
from pyqtorch.utils import (
    ATOL,
    RTOL,
    DensityMatrix,
    SolverType,
    density_mat,
    is_normalized,
    operator_kron,
    overlap,
    product_state,
    random_state,
)

pi = torch.tensor(torch.pi)


def Hamiltonian(batch_size: int = 1) -> torch.Tensor:
    sigmaz = torch.diag(torch.tensor([1.0, -1.0], dtype=DEFAULT_MATRIX_DTYPE))
    Hbase = torch.kron(sigmaz, sigmaz)
    H = torch.kron(Hbase, Hbase)
    if batch_size == 1:
        return H
    elif batch_size == 2:
        return torch.stack((H, H.conj()), dim=2)
    else:
        raise NotImplementedError


def Hamiltonian_general(n_qubits: int = 2, batch_size: int = 1) -> torch.Tensor:
    H_batch = torch.zeros(
        (2**n_qubits, 2**n_qubits, batch_size), dtype=DEFAULT_MATRIX_DTYPE
    )
    for i in range(batch_size):
        H_0 = torch.randn((2**n_qubits, 2**n_qubits), dtype=DEFAULT_MATRIX_DTYPE)
        H = (H_0 + torch.conj(H_0.transpose(0, 1))).to(DEFAULT_MATRIX_DTYPE)
        H_batch[..., i] = H
    return H_batch


def Hamiltonian_diag(n_qubits: int = 2, batch_size: int = 1) -> torch.Tensor:
    H_batch = torch.zeros(
        (2**n_qubits, 2**n_qubits, batch_size), dtype=DEFAULT_MATRIX_DTYPE
    )
    for i in range(batch_size):
        H_0 = torch.randn((2**n_qubits, 2**n_qubits), dtype=DEFAULT_MATRIX_DTYPE)
        H = H_0 + torch.conj(H_0.transpose(0, 1))
        get_diag = torch.diag(H)
        H_diag = torch.diag(get_diag)
        H_batch[..., i] = H_diag
    return H_batch


@pytest.mark.flaky(max_runs=5)
@pytest.mark.parametrize("n_qubits, batch_size", [(2, 1), (4, 2)])
def test_hamevo_general(n_qubits: int, batch_size: int) -> None:
    H = Hamiltonian_general(n_qubits, batch_size)
    t_evo = torch.rand(1, dtype=DEFAULT_REAL_DTYPE)
    hamevo = pyq.HamiltonianEvolution(H, t_evo, tuple([i for i in range(n_qubits)]))
    psi = random_state(n_qubits, batch_size)
    psi_star = hamevo(psi)
    assert is_normalized(psi_star, atol=ATOL)


@pytest.mark.flaky(max_runs=5)
def test_hamevo_single() -> None:
    n_qubits = 4
    H = Hamiltonian(1)
    t_evo = torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE)
    hamevo = pyq.HamiltonianEvolution(H, t_evo, tuple([i for i in range(n_qubits)]))
    psi = pyq.uniform_state(n_qubits)
    psi_star = hamevo(psi)
    result = overlap(psi_star, psi)
    assert torch.isclose(
        result, torch.tensor([0.5], dtype=torch.float64), rtol=RTOL, atol=ATOL
    )


@pytest.mark.flaky(max_runs=5)
def test_hamevo_batch() -> None:
    n_qubits = 4
    batch_size = 2
    H = Hamiltonian(batch_size)
    t_evo = torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE)
    hamevo = pyq.HamiltonianEvolution(H, t_evo, tuple([i for i in range(n_qubits)]))
    psi = pyq.uniform_state(n_qubits, batch_size)
    psi_star = hamevo(psi)
    result = overlap(psi_star, psi)
    assert torch.allclose(
        result, torch.tensor([0.5, 0.5], dtype=torch.float64), rtol=RTOL, atol=ATOL
    )


@pytest.mark.parametrize(
    "H, t_evo, target, batch_size",
    [
        (  # batchsize 1 | 1
            Hamiltonian(1),
            torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE),
            torch.tensor([0.5], dtype=torch.float64),
            1,
        ),
        (  # batchsize 1 | 4
            Hamiltonian(1),
            torch.tensor([torch.pi / 4, 0.0, torch.pi / 2, torch.pi]),
            torch.tensor([0.5, 1.0, 0.0, 1.0], dtype=torch.float64),
            4,
        ),
        (  # batchsize 2 | 1
            Hamiltonian(2),
            torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE),
            torch.tensor([0.5, 0.5], dtype=torch.float64),
            2,
        ),
        (  # batchsize 2 | 2
            Hamiltonian(2),
            torch.tensor([torch.pi / 4, torch.pi]),
            torch.tensor([0.5, 1.0], dtype=torch.float64),
            2,
        ),
    ],
)
def test_hamiltonianevolution_with_types(
    H: torch.Tensor,
    t_evo: torch.Tensor,
    target: torch.Tensor,
    batch_size: int,
) -> None:
    n_qubits = 4
    hamevo = pyq.HamiltonianEvolution(H, t_evo, tuple([i for i in range(n_qubits)]))
    assert len(hamevo._cache_hamiltonian_evo) == 0
    psi = pyq.uniform_state(n_qubits)
    psi_star = hamevo(psi)
    result = overlap(psi_star, psi)
    assert result.size() == (batch_size,)
    assert torch.allclose(result, target, rtol=RTOL, atol=ATOL)

    # test cached
    assert len(hamevo._cache_hamiltonian_evo) == 1

    psi_star = hamevo(psi)
    assert len(hamevo._cache_hamiltonian_evo) == 1
    result = overlap(psi_star, psi)
    # check value has not changed
    assert torch.allclose(result, target, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize(
    "H, t_evo, expected_state",
    [
        (
            Hamiltonian(1),
            torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE),
            torch.tensor([0.5], dtype=torch.float64),
        ),
        (
            Hamiltonian(1),
            torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE),
            torch.tensor([0.5, 0.5], dtype=torch.float64),
        ),
        (
            Hamiltonian(1),
            torch.tensor([torch.pi / 4, torch.pi]),
            torch.tensor([0.5, 1.0], dtype=torch.float64),
        ),
        # with batchdim
        (
            Hamiltonian(2),
            torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE),
            torch.tensor([0.5, 0.5], dtype=torch.float64),
        ),
        # with batchdim in position 0
        (
            Hamiltonian(2).transpose(0, 2),
            torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE),
            torch.tensor([0.5, 0.5], dtype=torch.float64),
        ),
    ],
)
def test_symbol_hamevo(
    H: torch.Tensor,
    t_evo: torch.Tensor,
    expected_state: torch.Tensor,
) -> None:
    symbol = "h"
    n_qubits = 4
    hamevo = pyq.HamiltonianEvolution(
        symbol, t_evo, tuple([i for i in range(n_qubits)])
    )
    assert hamevo.generator_type == GeneratorType.SYMBOL
    psi = pyq.uniform_state(n_qubits)
    psi_star = hamevo(psi, {symbol: H})
    state = overlap(psi_star, psi)
    assert torch.allclose(state, expected_state, rtol=RTOL, atol=ATOL)


def test_hamevo_fixed_tensor_result() -> None:
    hermitian_matrix = torch.tensor([[2.0, 1.0], [1.0, 3.0]], dtype=torch.complex128)
    hamevo = pyq.HamiltonianEvolution(
        generator=hermitian_matrix, time=torch.tensor([1.0j]), qubit_support=(0,)
    )
    assert hamevo.generator_type == GeneratorType.TENSOR
    expected_evo_result = torch.tensor(
        [[[13.1815 + 0.0j], [14.8839 + 0.0j]], [[14.8839 + 0.0j], [28.0655 + 0.0j]]],
        dtype=torch.complex128,
    )
    assert torch.allclose(hamevo.tensor(), expected_evo_result, atol=1.0e-4)


@pytest.mark.parametrize(
    "state_fn", [pyq.random_state, pyq.zero_state, pyq.uniform_state]
)
def test_parametric_phase_hamevo(
    state_fn: Callable, batch_size: int = 1, n_qubits: int = 1
) -> None:
    target = 0
    state = state_fn(n_qubits, batch_size=batch_size)
    phi = torch.rand(1, dtype=DEFAULT_MATRIX_DTYPE)
    H = (ZMAT - IMAT) / 2
    hamevo = pyq.HamiltonianEvolution(H, phi, (target,))
    assert hamevo.generator_type == GeneratorType.TENSOR
    phase = pyq.PHASE(target, "phi")
    assert torch.allclose(phase(state, {"phi": phi}), hamevo(state))


def test_hamevo_endianness() -> None:
    t = torch.ones(1)
    h = torch.tensor(
        [
            [0.9701, 0.0000, 0.7078, 0.0000],
            [0.0000, 0.9701, 0.0000, 0.7078],
            [0.4594, 0.0000, 0.9207, 0.0000],
            [0.0000, 0.4594, 0.0000, 0.9207],
        ],
        dtype=torch.complex128,
    )
    iszero = torch.tensor([False, True, False, True])
    op = pyq.HamiltonianEvolution(qubit_support=(0, 1), generator=h, time=t)
    st = op(pyq.zero_state(2)).flatten()
    assert torch.allclose(
        st[iszero], torch.zeros(1, dtype=DEFAULT_MATRIX_DTYPE), rtol=RTOL, atol=ATOL
    )

    h = torch.tensor(
        [
            [0.9701, 0.7078, 0.0000, 0.0000],
            [0.4594, 0.9207, 0.0000, 0.0000],
            [0.0000, 0.0000, 0.9701, 0.7078],
            [0.0000, 0.0000, 0.4594, 0.9207],
        ],
        dtype=torch.complex128,
    )
    iszero = torch.tensor([False, False, True, True])
    op = pyq.HamiltonianEvolution(qubit_support=(0, 1), generator=h, time=t)
    st = op(pyq.zero_state(2)).flatten()
    assert torch.allclose(
        st[iszero], torch.zeros(1, dtype=DEFAULT_MATRIX_DTYPE), rtol=RTOL, atol=ATOL
    )


def test_hamevo_endianness_cnot() -> None:
    n_qubits = 2
    state_10 = product_state("10")

    gen = -0.5 * operator_kron((IMAT - ZMAT).unsqueeze(-1), (IMAT - XMAT).unsqueeze(-1))
    hamiltonian_evolution = pyq.HamiltonianEvolution(
        generator=gen,
        time=torch.tensor([torch.pi / 2.0]),
        qubit_support=tuple(range(n_qubits)),
    )
    wf_hamevo = hamiltonian_evolution(state_10)

    cnot_op = pyq.CNOT(0, 1)
    wf_cnot = cnot_op(state_10)
    assert torch.allclose(wf_cnot, wf_hamevo, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("duration", [torch.rand(1), "duration"])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("use_sparse", [True, False])
@pytest.mark.parametrize("ode_solver", [SolverType.DP5_SE, SolverType.KRYLOV_SE])
def test_timedependent(
    tparam: str,
    param_y: float,
    duration: float,
    batch_size: int,
    use_sparse: bool,
    n_steps: int,
    torch_hamiltonian: Callable,
    hamevo_generator: Sequence,
    sin: tuple,
    sq: tuple,
    ode_solver: SolverType,
) -> None:

    psi_start = random_state(2, batch_size)

    dur_val = duration if isinstance(duration, torch.Tensor) else torch.rand(1)

    # simulate with time-dependent solver
    t_points = torch.linspace(0, dur_val[0], n_steps)

    psi_solver = pyq.sesolve(
        torch_hamiltonian,
        psi_start.reshape(-1, batch_size),
        t_points,
        ode_solver,
        options={"use_sparse": use_sparse},
    ).states[-1]

    # simulate with HamiltonianEvolution
    embedding = pyq.Embedding(
        tparam_name=tparam,
        var_to_call={sin[0]: sin[1], sq[0]: sq[1]},
    )
    hamiltonian_evolution = pyq.HamiltonianEvolution(
        generator=hamevo_generator,
        time=tparam,
        duration=dur_val,
        steps=n_steps,
        solver=ode_solver,
        use_sparse=use_sparse,
    )
    values = {"y": param_y, "duration": dur_val}
    psi_hamevo = hamiltonian_evolution(
        state=psi_start, values=values, embedding=embedding
    ).reshape(-1, batch_size)

    assert torch.allclose(psi_solver, psi_hamevo, rtol=RTOL, atol=1.0e-3)


@pytest.mark.parametrize("duration", [torch.rand(1), "duration"])
@pytest.mark.parametrize(
    "batchsize",
    [
        1,
        3,
    ],
)
def test_timedependent_with_noise(
    tparam: str,
    param_y: float,
    duration: float,
    batchsize: int,
    n_steps: int,
    torch_hamiltonian: Callable,
    hamevo_generator: Sequence,
    sin: tuple,
    sq: tuple,
) -> None:

    psi_start = density_mat(random_state(2, batchsize))
    dur_val = duration if isinstance(duration, torch.Tensor) else torch.rand(1)

    # simulate with time-dependent solver
    t_points = torch.linspace(0, dur_val[0], n_steps)

    # Define jump operators
    # Note that we squeeze to remove the batch dimension
    noise = AnalogDepolarizing(error_param=0.1, qubit_support=0)
    list_ops = noise._noise_operators(full_support=(0, 1))
    solver = SolverType.DP5_ME
    psi_solver = pyq.mesolve(
        torch_hamiltonian, psi_start, list_ops, t_points, solver
    ).states[-1]

    # simulate with HamiltonianEvolution
    embedding = pyq.Embedding(
        tparam_name=tparam,
        var_to_call={sin[0]: sin[1], sq[0]: sq[1]},
    )
    hamiltonian_evolution = pyq.HamiltonianEvolution(
        generator=hamevo_generator,
        time=tparam,
        duration=duration,
        steps=n_steps,
        solver=solver,
        noise=list_ops,
    )
    values = {"y": param_y, "duration": dur_val}
    psi_hamevo = hamiltonian_evolution(
        state=psi_start, values=values, embedding=embedding
    )
    assert isinstance(psi_hamevo, DensityMatrix)
    assert torch.allclose(psi_solver, psi_hamevo, rtol=RTOL, atol=1.0e-3)


def test_error_noise_qubit_support(
    tparam: str,
    duration: float,
    n_steps: int,
    hamevo_generator: Sequence,
):
    solver = SolverType.DP5_ME
    with pytest.raises(ValueError):
        noise = AnalogDepolarizing(error_param=0.1, qubit_support=3)
        hamevo = pyq.HamiltonianEvolution(
            generator=hamevo_generator,
            time=tparam,
            duration=duration,
            steps=n_steps,
            solver=solver,
            noise=noise,
        )


@pytest.mark.parametrize("n_qubits", [2, 4, 6])
@pytest.mark.parametrize("batch_size", [1, 2])
@pytest.mark.parametrize("make_param", [True, False])
@pytest.mark.parametrize("commuting_terms", [False, True])
def test_hamevo_parametric_gen(
    n_qubits: int, batch_size: int, make_param: bool, commuting_terms: bool
) -> None:
    k_1q = 2 * n_qubits  # Number of 1-qubit terms
    k_2q = n_qubits**2  # Number of 2-qubit terms
    generator, param_list = random_pauli_hamiltonian(
        n_qubits,
        k_1q,
        k_2q,
        make_param=make_param,
        p_param=1.0,
        commuting_terms=commuting_terms,
    )
    tparam = "t"

    param_list.append("t")

    hamevo = pyq.HamiltonianEvolution(generator, tparam, cache_length=2)

    if make_param:
        assert hamevo.generator_type in (
            GeneratorType.PARAMETRIC_OPERATION,
            GeneratorType.PARAMETRIC_COMMUTING_SEQUENCE,
        )
    else:
        assert hamevo.generator_type in (
            GeneratorType.OPERATION,
            GeneratorType.COMMUTING_SEQUENCE,
        )
    assert len(hamevo._cache_hamiltonian_evo) == 0

    def apply_hamevo_and_compare_expected(psi, values):
        psi_star = hamevo(psi, values)
        psi_expected = calc_mat_vec_wavefunction(hamevo, psi, values)
        assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)

    values = {param: torch.rand(batch_size) for param in param_list}

    psi = random_state(n_qubits)
    apply_hamevo_and_compare_expected(psi, values)

    # test cached
    if not commuting_terms:
        assert len(hamevo._cache_hamiltonian_evo) == 1
    apply_hamevo_and_compare_expected(psi, values)
    if not commuting_terms:
        assert len(hamevo._cache_hamiltonian_evo) == 1

    # test caching new value
    for param in param_list:
        values[param] += 0.1

    apply_hamevo_and_compare_expected(psi, values)
    if not commuting_terms:
        assert len(hamevo._cache_hamiltonian_evo) == 2

    # changing input state should not change the cache
    psi = random_state(n_qubits)
    apply_hamevo_and_compare_expected(psi, values)
    if not commuting_terms:
        assert len(hamevo._cache_hamiltonian_evo) == 2

    # test limit cache
    previous_cache_keys = hamevo._cache_hamiltonian_evo.keys()

    for param in param_list:
        values[param] += 0.1

    values_cache_key = str(OrderedDict(values))
    if not commuting_terms:
        assert values_cache_key not in previous_cache_keys

    apply_hamevo_and_compare_expected(psi, values)
    if not commuting_terms:
        assert len(hamevo._cache_hamiltonian_evo) == 2
        assert values_cache_key in previous_cache_keys


@pytest.mark.parametrize(
    "generator, time_param, result",
    [
        (RX(0, "x"), "x", True),
        (RX(1, 0.5), "y", False),
        (RX(0, "x"), "y", False),
        (RX(0, "x"), torch.tensor(0.5), False),
        (RX(0, torch.tensor(0.5)), torch.tensor(0.5), False),
        (Scale(X(1), "y"), "y", True),
        (Scale(X(1), 0.2), "x", False),
        (
            Add(
                [Scale(X(1), ConcretizedCallable("mul", ["y", "x"])), Scale(X(1), "z")]
            ),
            "x",
            True,
        ),
        (
            Add(
                [Scale(X(1), ConcretizedCallable("add", ["y", "x"])), Scale(X(1), "z")]
            ),
            "t",
            False,
        ),
    ],
)
def test_hamevo_is_time_dependent_generator(generator, time_param, result) -> None:
    hamevo = HamiltonianEvolution(generator, time_param)
    assert hamevo.is_time_dependent == result



from __future__ import annotations

from torch import Tensor


def mutate_separate_target(
    state: Tensor, target_qubit: tuple[int, ...]
) -> tuple[list[int], Tensor]:
    """Create a tensor separating the target components
    for a single-qubit gate for mutating an input state-vector.

    Args:
        state (Tensor): Input state.
        target_qubit (tuple[int, ...]): Target indices.

    Returns:
        tuple[int list[int], Tensor]: The permutation indices with an intermediate state
            with separated target qubit.
    """
    n_qubits = len(state.shape) - 1
    perm = list(range(n_qubits + 1))
    perm = list(target_qubit) + list(filter(lambda x: x not in target_qubit, perm))

    # Transpose the state
    state = state.permute(perm)

    # Reshape to separate the target qubit
    state = state.reshape(2 * len(target_qubit), -1)
    return perm, state


def mutate_revert_modified(
    state: Tensor, original_shape: tuple[int], perm: list[int]
) -> Tensor:
    """After mutating a state given a single qubit gate, we revert back the new state
    to correspond to the `original_shape`.

    Args:
        state (Tensor): modified state by operation.
        original_shape (tuple[int]): original shape for reshapping.
        perm (list[int]): Permutation indices.

    Returns:
        Tensor: Mutated state.
    """
    # Reshape back to original structure
    state = state.reshape(original_shape)

    # Transpose back to original order
    inverse_perm = [perm.index(i) for i in range(len(perm))]
    return state.permute(inverse_perm)


def mutate_control_slice(n_qubits: int, control_qubits: tuple[int, ...]) -> tuple:
    """Create a slice for mutable controlled operations.

    Args:
        n_qubits (int): Number of qubits
        control_qubits (tuple[int, ...]): Control indices.

    Returns:
        tuple: slices to select controlled states.
    """
    selector: list = [slice(None)] * n_qubits
    for qubit in control_qubits:
        selector[qubit] = 1
    return tuple(selector)



from __future__ import annotations

import logging
from collections import OrderedDict
from functools import cached_property
from logging import getLogger
from typing import Callable, Tuple, Union
from uuid import uuid4

import torch
from torch import Tensor
from torch.nn import ModuleList, ParameterDict

from pyqtorch.apply import apply_operator
from pyqtorch.circuit import Sequence
from pyqtorch.composite import Add, Scale
from pyqtorch.embed import ConcretizedCallable, Embedding
from pyqtorch.noise import AnalogNoise
from pyqtorch.primitives import Primitive
from pyqtorch.quantum_operation import QuantumOperation
from pyqtorch.time_dependent.sesolve import sesolve
from pyqtorch.utils import (
    Operator,
    SolverType,
    State,
    StrEnum,
    _round_operator,
    expand_operator,
    finitediff,
    is_diag_batched,
    is_parametric,
)

BATCH_DIM = 2
TGenerator = Union[Tensor, str, QuantumOperation, Sequence]

logger = getLogger(__name__)


def forward_hook(*args, **kwargs) -> None:  # type: ignore[no-untyped-def]
    logger.debug("Forward complete")
    torch.cuda.nvtx.range_pop()


def pre_forward_hook(*args, **kwargs) -> None:  # type: ignore[no-untyped-def]
    logger.debug("Executing forward")
    torch.cuda.nvtx.range_push("HamiltonianEvolution.forward")


def backward_hook(*args, **kwargs) -> None:  # type: ignore[no-untyped-def]
    logger.debug("Backward complete")
    torch.cuda.nvtx.range_pop()


def pre_backward_hook(*args, **kwargs) -> None:  # type: ignore[no-untyped-def]
    logger.debug("Executed backward")
    torch.cuda.nvtx.range_push("Hamiltonian Evolution.backward")


class GeneratorType(StrEnum):
    """
    Options for types of generators allowed in HamiltonianEvolution.
    """

    PARAMETRIC_OPERATION = "parametric_operation"
    """Generators of type Primitive or Sequence which contain
       possibly trainable or non-trainable parameters."""
    OPERATION = "operation"
    """Generators of type Primitive or Sequence which do not contain parameters or contain
       constants as Parameters for example pyq.Scale(Z(0), torch.tensor([1.]))."""
    PARAMETRIC_COMMUTING_SEQUENCE = "parametric_commuting_sequence"
    """Parametric generators of type Add where each operation commute with each other."""
    COMMUTING_SEQUENCE = "commuting_sequence"
    """Generators of type Add where each operation commute with each other."""
    TENSOR = "tensor"
    """Generators of type torch.Tensor in which case a qubit_support needs to be passed."""
    SYMBOL = "symbol"
    """Generators which are symbolic, i.e. will be passed via the 'values' dict by the user."""


COMMUTING = (
    GeneratorType.PARAMETRIC_COMMUTING_SEQUENCE,
    GeneratorType.COMMUTING_SEQUENCE,
)


def evolve(
    hamiltonian: Operator, time_evolution: Tensor, diagonal: bool = False
) -> Operator:
    """Get the evolved operator.

    For a hamiltonian :math:`H` and a time evolution :math:`t`, returns :math:`exp(-i H, t)`

    Arguments:
        hamiltonian: The operator :math:`H` for evolution.
        time_evolution: The evolution time :math:`t`.
        diagonal: Whether the hamiltonian is diagonal or not.
            Note for diagonal cases, we return a dense tensor.

    Returns:
        The evolution operator.
    """
    if diagonal:
        evol_operator = torch.transpose(hamiltonian, 0, 1) * (
            -1j * time_evolution
        ).view((-1, 1))
        evol_operator = torch.diag_embed(torch.exp(evol_operator))
    # for case 3D tensor is passed as generator
    elif is_diag_batched(hamiltonian):
        evol_operator = torch.diagonal(hamiltonian) * (-1j * time_evolution).view(
            (-1, 1)
        )
        evol_operator = torch.diag_embed(torch.exp(evol_operator))
    else:
        evol_operator = torch.transpose(hamiltonian, 0, -1) * (
            -1j * time_evolution
        ).view((-1, 1, 1))
        evol_operator = torch.linalg.matrix_exp(evol_operator)
    return torch.transpose(evol_operator, 0, -1)


class HamiltonianEvolution(Sequence):
    """
    The HamiltonianEvolution corresponds to :math:`t`, returns :math:`exp(-i H, t)` where
    a hamiltonian/generator :math:`H` and a time evolution :math:`t` are given.

    Note that the quantity :math:`H.t` is considered dimensionless.

    We can create such operation by passing different generator types:
        - A tensor representation of the generator,
        - A string when we consider the generator as a symbol.
        - Operations as a single primitive or a sequence, parameterized or not.

    Attributes:
        generator_type: This attribute informs on how the generator is inputed
        and sets the logic for applying hamiltonian evolution.
        time: The evolution time :math:`t`.
        operations: List of operations.
        cache_length: LRU cache cache_length evolution operators for given set
                    of parameter values.
        duration: Total duration for evolving when using a solver.
        steps: Number of steps to use when using solver.
        solver: Time-dependent Lindblad master equation solver.
        noise_operators: List of tensors or Kraus operators adding analog noise
            when solving with a Shrodinger equation solver.
    """

    def __init__(
        self,
        generator: TGenerator,
        time: Tensor | str | ConcretizedCallable,
        qubit_support: Tuple[int, ...] | None = None,
        cache_length: int = 1,
        duration: Tensor | str | float | None = None,
        steps: int = 100,
        solver: SolverType = SolverType.DP5_SE,
        use_sparse: bool = False,
        noise: list[Tensor] | AnalogNoise | None = None,
    ):
        """Initializes the HamiltonianEvolution.
        Depending on the generator argument, set the type and set the right generator getter.

        Arguments:
            generator: The generator :math:`H`.
            time: The evolution time :math:`t`.
            qubit_support: The qubits the operator acts on. If generator is a quantum
                operation or sequence of operations,
                it will be inferred from the generator.
            cache_length: LRU cache cache_length evolution operators for given set
                    of parameter values.
            duration: Total duration for evolving when using a solver.
            steps: Number of steps to use when using solver.
            solver: Time-dependent Lindblad master equation solver.
            noise: List of jump operatoes for noisy simulations or an AnalogNoise
                when solving with a Shrodinger equation solver.
        """

        self.solver_type = solver
        self.steps = steps
        self.duration = duration
        self.use_sparse = use_sparse
        self.is_diagonal = False
        self._param_uuid = str(uuid4())
        original_generator = None

        if isinstance(duration, (str, float, Tensor)) or duration is None:
            self.duration = duration
        else:
            raise ValueError(
                "Optional argument `duration` should be passed as str, float or Tensor."
            )

        if isinstance(time, (str, Tensor, ConcretizedCallable)):
            self.time = time
        else:
            raise ValueError(
                "Argument `time` should be passed as str, Tensor or ConcretizedCallable."
            )

        if isinstance(generator, Tensor):
            if qubit_support is None:
                raise ValueError(
                    "When using a Tensor generator, please pass a qubit_support."
                )
            if len(generator.shape) < 3:
                generator = generator.unsqueeze(2)
            generator = [Primitive(generator, qubit_support)]
            self.is_diagonal = generator[0].is_diagonal
            self.generator_type = GeneratorType.TENSOR

        elif isinstance(generator, str):
            if qubit_support is None:
                raise ValueError(
                    "When using a symbolic generator, please pass a qubit_support."
                )
            self.generator_type = GeneratorType.SYMBOL
            self.generator_symbol = generator
            generator = []
        elif (
            noise is None and isinstance(generator, Add) and generator.commuting_terms()
        ):
            qubit_support = generator.qubit_support
            self.generator_type = (
                GeneratorType.PARAMETRIC_COMMUTING_SEQUENCE
                if is_parametric(generator)
                else GeneratorType.COMMUTING_SEQUENCE
            )
            original_generator = generator
            generator = [
                HamiltonianEvolution(
                    op,
                    time,
                    cache_length=cache_length,
                    duration=duration,
                    steps=steps,
                    solver=solver,
                    use_sparse=use_sparse,
                    noise=noise,
                )
                for op in generator.operations
            ]
            for gen in generator:
                gen._param_uuid = self._param_uuid
            self.is_diagonal = all(gen.is_diagonal for gen in generator)

        elif isinstance(generator, (QuantumOperation, Sequence)):
            qubit_support = generator.qubit_support

            if is_parametric(generator):
                generator = [generator]
                self.generator_type = GeneratorType.PARAMETRIC_OPERATION
            else:
                # avoiding using dense tensor for diagonal generators
                tgen = generator.tensor(diagonal=generator.is_diagonal)
                generator = [
                    Primitive(
                        tgen, generator.qubit_support, diagonal=(len(tgen.size()) == 2)
                    )
                ]
                self.is_diagonal = generator[0].is_diagonal
                self.generator_type = GeneratorType.OPERATION
        else:
            raise TypeError(
                f"Received generator of type {type(generator)},\
                            allowed types are: [Tensor, str, Primitive, Sequence]"
            )
        super().__init__(generator)
        self._qubit_support = qubit_support  # type: ignore
        self._original_generator = original_generator

        logger.debug("Hamiltonian Evolution initialized")
        if logger.isEnabledFor(logging.DEBUG):
            # When Debugging let's add logging and NVTX markers
            # WARNING: incurs performance penalty
            self.register_forward_hook(forward_hook, always_call=True)
            self.register_full_backward_hook(backward_hook)
            self.register_forward_pre_hook(pre_forward_hook)
            self.register_full_backward_pre_hook(pre_backward_hook)

        self._generator_map: dict[GeneratorType, Callable] = {
            GeneratorType.SYMBOL: self._symbolic_generator,
            GeneratorType.TENSOR: self._generator,
            GeneratorType.OPERATION: self._generator,
            GeneratorType.PARAMETRIC_OPERATION: self._generator,
            GeneratorType.PARAMETRIC_COMMUTING_SEQUENCE: self._commuting_generator,
            GeneratorType.COMMUTING_SEQUENCE: self._commuting_generator,
        }

        # to avoid recomputing hamiltonians and evolution
        self._cache_hamiltonian_evo: dict[str, Tensor] = dict()
        self.cache_length = cache_length

        if isinstance(noise, list):
            noise = AnalogNoise(noise, self.qubit_support)
        if noise is not None and set(noise.qubit_support) - set(self.qubit_support):
            raise ValueError(
                "The noise should be a subset or the same qubit support"
                "as HamiltonianEvolution."
            )
        self.noise = noise

    @property
    def generator(self) -> ModuleList:
        """Returns the operations making the generator.

        Returns:
            The generator as a ModuleList.
        """
        return self.operations

    @property
    def is_parametric(self) -> bool:
        """Check if operation is parametric, that only the time is parametrized.


        Returns:
            bool: True is `time` is str.
        """
        return isinstance(self.param_name, str)

    @property
    def is_parametric_generator(self) -> bool:
        return self.generator_type in (
            GeneratorType.SYMBOL,
            GeneratorType.PARAMETRIC_OPERATION,
            GeneratorType.PARAMETRIC_COMMUTING_SEQUENCE,
        )

    @cached_property
    def is_time_dependent(self) -> bool:
        return self._is_time_dependent(self.generator)

    def _flatten(self) -> ModuleList:
        return ModuleList([self])

    def flatten(self) -> ModuleList:
        return self._flatten()

    @property
    def param_name(self) -> Tensor | str:
        return self.time

    def _is_time_dependent(self, generator: TGenerator) -> bool:
        from pyqtorch.primitives import Parametric

        res = False
        if isinstance(self.time, Tensor):
            return res
        else:
            if isinstance(generator, (Sequence, QuantumOperation, ModuleList)):
                for m in generator.modules():
                    if isinstance(m, (Scale, Parametric)):
                        if self.time in getattr(m.param_name, "independent_args", []):
                            # param_name is a ConcretizedCallable object
                            res = True
                        elif m.param_name == self.time:
                            # param_name is a string
                            res = True
        return res

    def _symbolic_generator(
        self,
        values: dict,
        embedding: Embedding | None = None,
        full_support: tuple[int, ...] | None = None,
    ) -> Operator:
        """Returns the generator for the SYMBOL case.

        Arguments:
            values: Values dict with any needed parameters.
            embedding: Embedding of parameters.
            full_support: The qubits the returned tensor
                will be defined over. Defaults to None for only using the qubit_support.

        Returns:
            The generator as a tensor.
        """
        hamiltonian = values[self.generator_symbol]
        # add batch dim
        if len(hamiltonian.shape) == 2:
            return hamiltonian.unsqueeze(2)
        # cases when the batchdim is at index 0 instead of 2
        if len(hamiltonian.shape) == 3 and (
            hamiltonian.shape[0] != hamiltonian.shape[1]
        ):
            return torch.transpose(hamiltonian, 0, 2)
        if len(hamiltonian.shape) == 4 and (
            hamiltonian.shape[0] != hamiltonian.shape[1]
        ):
            return torch.permute(hamiltonian.squeeze(3), (1, 2, 0))
        return hamiltonian

    def _generator(
        self,
        values: dict = dict(),
        embedding: Embedding | None = None,
        full_support: tuple[int, ...] | None = None,
    ) -> Operator:
        """Returns the TENSOR, OPERATION and PARAMETRIC_OPERATION generator.

        Arguments:
            values: Values dict with any needed parameters.
            embedding: Embedding of parameters.
            full_support: The qubits the returned tensor
                will be defined over. Defaults to None for only using the qubit_support.

        Returns:
            The generator as a tensor.
        """
        return super().tensor(
            values, embedding, full_support, diagonal=self.is_diagonal
        )

    def _commuting_generator(
        self,
        values: dict = dict(),
        embedding: Embedding | None = None,
        full_support: tuple[int, ...] | None = None,
    ) -> Operator:
        """Returns the COMMUTING_SEQUENCE generator.

        Arguments:
            values: Values dict with any needed parameters.
            embedding: Embedding of parameters.
            full_support: The qubits the returned tensor
                will be defined over. Defaults to None for only using the qubit_support.

        Returns:
            The generator as a tensor.
        """
        return self._original_generator.tensor(  # type: ignore [union-attr]
            values, embedding, full_support, diagonal=self.is_diagonal
        )

    @property
    def create_hamiltonian(self) -> Callable:
        """A utility method for setting the right generator getter depending on the init case.

        Returns:
            The right generator getter.
        """
        return self._generator_map[self.generator_type]

    @cached_property
    def eigenvals_generator(self) -> Tensor:
        """Get eigenvalues of the underlying hamiltonian.

        Note: Only works for GeneratorType.TENSOR
        or GeneratorType.OPERATION.

        Returns:
            Eigenvalues of the operation.
        """
        if self.generator_type not in COMMUTING:
            return self.generator[0].eigenvalues
        else:
            blockmat = self.create_hamiltonian()
            if len(blockmat.shape) == 3:
                return torch.linalg.eigvals(blockmat.permute((2, 0, 1))).reshape(-1, 1)
            else:
                # for diagonal cases
                return blockmat

    @cached_property
    def spectral_gap(self) -> Tensor:
        """Difference between the moduli of the two largest eigenvalues of the generator.

        Returns:
            Tensor: Spectral gap value.
        """
        spectrum = self.eigenvals_generator
        diffs = spectrum - spectrum.T
        diffs = _round_operator(diffs)
        spectral_gap = torch.unique(torch.abs(torch.tril(diffs)))
        return spectral_gap[spectral_gap.nonzero()]

    def _time_evolution(self, values: dict[str, Tensor] | ParameterDict) -> Tensor:
        """Get the evolution from parameter values.

        Arguments:
            values: Values of parameters.

        Returns:
            The time evolution.
        """
        if isinstance(self.time, str):
            # note: GPSR trick when the same param_name is used in many operations
            if self._param_uuid in values.keys():
                time_evolution = values[self.time] + values[self._param_uuid]
            else:
                time_evolution = values[self.time]
        elif isinstance(self.time, ConcretizedCallable):
            time_evolution = self.time(values)
        else:
            time_evolution = self.time
        return time_evolution

    def _forward(
        self,
        state: Tensor,
        values: dict[str, Tensor] | ParameterDict = dict(),
        embedding: Embedding | None = None,
    ) -> State:
        if self.generator_type not in COMMUTING:

            evolved_op = self.tensor(values, embedding)
            return apply_operator(
                state=state, operator=evolved_op, qubit_support=self.qubit_support
            )
        else:
            return self._forward_commuting_generators(state, values, embedding)

    def _forward_commuting_generators(
        self,
        state: Tensor,
        values: dict[str, Tensor] | ParameterDict = dict(),
        embedding: Embedding | None = None,
    ) -> State:
        """Apply the hamiltonian evolution with input parameter values when
        hamiltonian is composed of commuting terms.

        Arguments:
            state: Input state.
            values: Values of parameters.
            embedding: Embedding of parameters.

        Returns:
            The transformed state.
        """
        if embedding is not None:
            values = embedding(values)

        return super().forward(state, values, embedding)

    def _forward_time(
        self,
        state: Tensor,
        duration: float,
        values: dict[str, Tensor] | ParameterDict = dict(),
        embedding: Embedding = Embedding(),
    ) -> State:
        """Apply the hamiltonian evolution with input parameter values for time dependent cases.

        Arguments:
            state: Input state.
            values: Values of parameters.
            embedding: Embedding of parameters.

        Returns:
            The transformed state.
        """
        n_qubits = len(state.shape) - 1
        batch_size = state.shape[-1]
        t_grid = torch.linspace(0, duration, self.steps)
        if embedding is not None:
            values.update({embedding.tparam_name: torch.tensor(0.0)})  # type: ignore [dict-item]
            embedded_params = embedding(values)
        else:
            embedded_params = values

        def Ht(t: torch.Tensor) -> torch.Tensor:
            """Accepts a value 't' for time and returns
            a (2**n_qubits, 2**n_qubits) Hamiltonian evaluated at time 't'.
            """
            # We use the original embedded params and return a new dict
            # where we reembedded all parameters depending on time with value 't'
            if embedding is not None:
                reembedded_time_values = embedding.reembed_tparam(
                    embedded_params, torch.as_tensor(t)
                )
            else:
                values[self.time] = torch.as_tensor(t)
                reembedded_time_values = values
            return self.create_hamiltonian(
                reembedded_time_values,
                embedding,
                full_support=tuple(range(n_qubits)),
            ).squeeze(2)

        if self.noise is None:
            sol = sesolve(
                Ht,
                torch.flatten(state, start_dim=0, end_dim=-2),
                t_grid,
                self.solver_type,
                options={"use_sparse": self.use_sparse},
            )

            # Retrieve the last state of shape (2**n_qubits, batch_size)
            # and reshape
            state = sol.states[-1].reshape([2] * n_qubits + [batch_size])
        else:
            state = self.noise(
                state,
                Ht,
                t_grid,
                self.solver_type,
                options={"use_sparse": self.use_sparse},
                full_support=self.qubit_support,
            )
        return state

    def forward(
        self,
        state: Tensor,
        values: dict[str, Tensor] | ParameterDict | None = None,
        embedding: Embedding | None = None,
    ) -> State:
        """Apply the hamiltonian evolution with input parameter values.

        Note when duration is None for the time_dependent generator cases, we fall back
        to considering forwarding as a time-independent case.

        Arguments:
            state: Input state.
            values: Values of parameters.
            embedding: Embedding of parameters.

        Returns:
            The transformed state.
        """
        values = values or dict()
        if self.is_time_dependent or (
            embedding is not None and getattr(embedding, "tparam_name", None)
        ):
            # to handle cases where a time-dependent generator becomes
            # time-independent with missing duration
            duration = (
                values[self.duration]
                if isinstance(self.duration, str)
                else self.duration
            )
            if duration is None:
                return self._forward(
                    state,
                    values,
                    embedding,
                )
            return self._forward_time(state, float(duration), values, embedding)  # type: ignore [arg-type]

        else:
            return self._forward(state, values, embedding)

    def tensor(
        self,
        values: dict | None = None,
        embedding: Embedding | None = None,
        full_support: tuple[int, ...] | None = None,
        diagonal: bool = False,
    ) -> Operator:
        """Get the corresponding unitary over n_qubits.

        To avoid computing the evolution operator, we store it in cache wrt values.

        Arguments:
            values: Parameter values.
            embedding (Embedding | None, optional): Optional embedding. Defaults to None.
            full_support (tuple[int, ...] | None, optional): The qubits the returned tensor
                will be defined over. Defaults to None for only using the qubit_support.
            diagonal (bool, optional): Whether to return the diagonal form of the tensor or not.
                Defaults to False.

        Returns:
            The unitary representation.
        """
        values = values or dict()
        if embedding is not None:
            values = embedding(values)
        use_diagonal = diagonal and self.is_diagonal

        values_cache_key = str(OrderedDict(values))
        commuting_generator = self.generator_type in COMMUTING
        if commuting_generator:
            return super().tensor(values, embedding, full_support, use_diagonal)
        elif self.cache_length > 0 and values_cache_key in self._cache_hamiltonian_evo:
            evolved_op = self._cache_hamiltonian_evo[values_cache_key]

        else:
            hamiltonian: torch.Tensor = self.create_hamiltonian(values, embedding)  # type: ignore [call-arg]
            time_evolution = self._time_evolution(values)

            evolved_op = evolve(hamiltonian, time_evolution, diagonal=self.is_diagonal)
            if use_diagonal:
                evolved_op = torch.diagonal(evolved_op).T
            nb_cached = len(self._cache_hamiltonian_evo)

            # LRU caching
            if (nb_cached > 0) and (nb_cached == self.cache_length):
                self._cache_hamiltonian_evo.pop(next(iter(self._cache_hamiltonian_evo)))
            if nb_cached < self.cache_length:
                self._cache_hamiltonian_evo[values_cache_key] = evolved_op

        if full_support is None:
            return evolved_op
        else:
            return expand_operator(
                evolved_op, self.qubit_support, full_support, use_diagonal
            )

    def jacobian(
        self,
        values: dict[str, Tensor] = dict(),
        embedding: Embedding | None = None,
    ) -> Tensor:
        """Get the corresponding unitary of the jacobian.

        Arguments:
            values: Parameter value.

        Returns:
            The unitary representation of the jacobian.
        """
        if embedding is not None:
            values = embedding(values)

        hamiltonian: torch.Tensor = self.create_hamiltonian(values, embedding)  # type: ignore [call-arg]

        if self._param_uuid in values.keys():
            # note: GPSR trick when the same param_name is used in many operations
            val = values[self.param_name] + values[self._param_uuid]
        else:
            val = values[self.param_name]
        return finitediff(
            lambda t: evolve(hamiltonian, t, self.is_diagonal),
            val.reshape(-1, 1),
            (0,),
        )



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Callable

import torch
from torch import Tensor

from pyqtorch.time_dependent.options import AdaptiveSolverOptions
from pyqtorch.utils import hairer_norm


class AdaptiveIntegrator:
    """Adaptive step-size ODE integrator.

    For details about the integration method, see Chapter II.4 of [1].

    [1] Hairer et al., Solving Ordinary Differential Equations I (1993), Springer
        Series in Computational Mathematics.
    """

    def __init__(
        self,
        H: Tensor | Callable,
        y0: Tensor,
        tsave: Tensor,
        options: AdaptiveSolverOptions,
    ):
        self.H = H
        self.t0 = 0.0
        self.y0 = y0
        self.tsave = tsave
        self.options = options

        # initialize the step counter
        self.step_counter = 0

    def init_forward(self) -> tuple:
        # initial values of the ODE routine
        f0 = self.ode_fun(self.t0, self.y0)
        dt0 = self.init_tstep(self.t0, self.y0, f0, self.ode_fun)
        error0 = 1.0
        return self.t0, self.y0, f0, dt0, error0

    @property
    @abstractmethod
    def order(self) -> int:
        pass

    @property
    @abstractmethod
    def tableau(self) -> tuple[Tensor, Tensor, Tensor, Tensor]:
        pass

    @abstractmethod
    def step(
        self,
        t0: float,
        y0: Tensor,
        f0: Tensor,
        dt: float,
        fun: Callable[[float, Tensor], Tensor],
    ) -> tuple[Tensor, Tensor, Tensor]:
        """Compute a single step of the ODE integration."""
        pass

    @abstractmethod
    def ode_fun(self, t: float, y: Tensor) -> Tensor:
        pass

    def integrate(self, t0: float, t1: float, y: Tensor, *args: Any) -> tuple:
        ft, dt, error = args

        cache = (dt, error)
        t = t0
        while t < t1:

            dt = self.update_tstep(dt, error)

            # check for time overflow
            if t + dt >= t1:
                cache = (dt, error)
                dt = t1 - t

            # compute the next step
            ft_new, y_new, y_err = self.step(t, y, ft, dt, self.ode_fun)

            error = self.get_error(y_err, y, y_new)
            if error <= 1:
                t, y, ft = t + dt, y_new, ft_new

            # check max steps are not reached
            self.increment_step_counter(t)

        dt, error = cache
        return y, ft, dt, error

    def increment_step_counter(self, t: float) -> None:
        """Increment the step counter and check for max steps."""
        self.step_counter += 1
        if self.step_counter == self.options.max_steps:
            raise RuntimeError(
                "Maximum number of time steps reached in adaptive time step ODE"
                f" solver at time t={t:.2g} (`max_steps={self.options.max_steps}`)."
                " This is likely due to a diverging solution. Try increasing the"
                " maximum number of steps, or use a different solver."
            )

    @torch.no_grad()
    def get_error(self, y_err: Tensor, y0: Tensor, y1: Tensor) -> float:
        """Compute the error of a given solution.

        See Equation (4.11) of [1].
        """
        scale = self.options.atol + self.options.rtol * torch.max(y0.abs(), y1.abs())
        return float(hairer_norm(y_err / scale).max().item())

    @torch.no_grad()
    def init_tstep(
        self, t0: float, y0: Tensor, f0: Tensor, fun: Callable[[float, Tensor], Tensor]
    ) -> float:
        """Initialize the time step of an adaptive step size integrator.

        See Equation (4.14) of [1] for the detailed steps. For this function, we keep
        the same notations as in the book.
        """

        sc = self.options.atol + torch.abs(y0) * self.options.rtol
        f0 = f0.to_dense() if self.options.use_sparse else f0
        d0, d1 = (
            hairer_norm(y0 / sc).max().item(),
            hairer_norm(f0 / sc).max().item(),
        )

        if d0 < 1e-5 or d1 < 1e-5:
            h0 = 1e-6
        else:
            h0 = 0.01 * d0 / d1

        y1 = y0 + h0 * f0
        f1 = fun(t0 + h0, y1)
        diff = (f1 - f0).to_dense() if self.options.use_sparse else f1 - f0
        d2 = hairer_norm(diff / sc).max().item() / h0
        if d1 <= 1e-15 and d2 <= 1e-15:
            h1 = max(1e-6, h0 * 1e-3)
        else:
            h1 = (0.01 / max(d1, d2)) ** (1.0 / float(self.order + 1))

        return min(100 * h0, h1)

    @torch.no_grad()
    def update_tstep(self, dt: float, error: float) -> float:
        """Update the time step of an adaptive step size integrator.

        See Equation (4.12) and (4.13) of [1] for the detailed steps.
        """
        if error == 0:  # no error -> maximally increase the time step
            return dt * self.options.max_factor

        elif error <= 1:  # time step accepted -> take next time step at least as large
            return float(
                dt
                * max(
                    1.0,
                    min(
                        self.options.max_factor,
                        self.options.safety_factor * error ** (-1.0 / self.order),
                    ),
                )
            )

        else:  # time step rejected -> reduce next time step
            return float(
                dt
                * max(
                    self.options.min_factor,
                    self.options.safety_factor * error ** (-1.0 / self.order),
                )
            )

    def run(self) -> Tensor:
        """Integrates the ODE forward from time `self.t0` to time `self.tstop[-1]`
        starting from initial state `self.y0`, and save the state for each time in
        `self.tstop`."""

        # initialize the ODE routine
        t, y, *args = self.init_forward()

        # run the ODE routine
        result = []
        for tnext in self.tsave:
            y, *args = self.integrate(t, tnext, y, *args)
            result.append(y)
            t = tnext

        if len(y.shape) == 2:
            res = torch.stack(result)
        elif len(y.shape) == 3:
            res = torch.stack(result).permute(0, 2, 3, 1)

        return res



from __future__ import annotations

import torch
from torch import Tensor

from pyqtorch.time_dependent.integrators.krylov import KrylovIntegrator


class Krylov(KrylovIntegrator):
    """Krylov subspace method for solving Schrodinger equation."""

    def integrate(self, t0: float, t1: float, y: Tensor) -> Tensor:
        dt = t1 - t0
        y = self.step(dt, t1, y)
        return y

    def step(
        self,
        dt: float,
        t: float,
        state: Tensor,
    ) -> Tensor:

        def exponentiate() -> tuple[torch.Tensor, bool]:
            # approximate next iteration by modifying T, and unmodifying
            T[i - 1, i] = 0
            T[i + 1, i] = 1
            exp = torch.linalg.matrix_exp(-1j * dt * T[: i + 2, : i + 2].clone())
            T[i - 1, i] = T[i, i - 1]
            T[i + 1, i] = 0

            e1 = abs(exp[i, 0])
            e2 = abs(exp[i + 1, 0]) * n
            if e1 > 10 * e2:
                error = e2
            elif e2 > e1:
                error = e1
            else:
                error = (e1 * e2) / (e1 - e2)

            converged = error < self.options.exp_tolerance
            return exp[:, 0], converged

        lanczos_vectors = [state]
        T = torch.zeros(
            self.options.max_krylov + 1, self.options.max_krylov + 1, dtype=state.dtype
        )

        # step 0 of the loop
        v = torch.matmul(self.H(t), state)
        a = torch.matmul(v.conj().T, state)
        n = torch.linalg.vector_norm(v)
        T[0, 0] = a
        v = v - a * state

        for i in range(1, self.options.max_krylov):
            # this block should not be executed in step 0
            b = torch.linalg.vector_norm(v)
            if b < self.options.norm_tolerance:
                exp = torch.linalg.matrix_exp(-1j * dt * T[:i, :i])
                weights = exp[:, 0]
                converged = True
                break

            T[i, i - 1] = b
            T[i - 1, i] = b
            state = v / b
            lanczos_vectors.append(state)
            weights, converged = exponentiate()
            if converged:
                break

            v = torch.matmul(self.H(t), state)
            a = torch.matmul(v.conj().T, state)
            n = torch.linalg.vector_norm(v)
            T[i, i] = a
            v = v - a * lanczos_vectors[i] - b * lanczos_vectors[i - 1]

        if not converged:
            raise RecursionError(
                "Exponentiation algorithm did not converge \
                to precision in allotted number of steps."
            )

        result = lanczos_vectors[0] * weights[0]
        for i in range(1, len(lanczos_vectors)):
            result += lanczos_vectors[i] * weights[i]

        return result



from __future__ import annotations

import functools
from typing import Any, Callable

import torch
from torch import Tensor

from pyqtorch.time_dependent.integrators.adaptive import AdaptiveIntegrator


class DormandPrince5(AdaptiveIntegrator):
    """Dormand-Prince method for adaptive time step ODE integration.

    This is a fifth order solver that uses a fourth order solution to estimate the
    integration error. It does so using only six function evaluations. See `Dormand and
    Prince, A family of embedded Runge-Kutta formulae (1980), Journal of Computational
    and Applied Mathematics`. See also `Shampine, Some Practical Runge-Kutta Formulas
    (1986), Mathematics of Computation`.
    """

    def __init__(self, *args: Any) -> None:
        super().__init__(*args)

    @functools.cached_property
    def order(self) -> int:
        return 5

    @functools.cached_property
    def tableau(self) -> tuple[Tensor, Tensor, Tensor, Tensor]:
        """Build the Butcher tableau of the integrator."""
        alpha = torch.tensor(
            [1 / 5, 3 / 10, 4 / 5, 8 / 9, 1.0, 1.0, 0.0], dtype=self.options.rtype
        )
        beta = torch.tensor(
            [
                [1 / 5, 0, 0, 0, 0, 0, 0],
                [3 / 40, 9 / 40, 0, 0, 0, 0, 0],
                [44 / 45, -56 / 15, 32 / 9, 0, 0, 0, 0],
                [19372 / 6561, -25360 / 2187, 64448 / 6561, -212 / 729, 0, 0, 0],
                [9017 / 3168, -355 / 33, 46732 / 5247, 49 / 176, -5103 / 18656, 0, 0],
                [35 / 384, 0, 500 / 1113, 125 / 192, -2187 / 6784, 11 / 84, 0],
            ],
            dtype=self.options.ctype,
        )
        csol5 = torch.tensor(
            [35 / 384, 0, 500 / 1113, 125 / 192, -2187 / 6784, 11 / 84, 0],
            dtype=self.options.ctype,
        )
        csol4 = torch.tensor(
            [
                5179 / 57600,
                0,
                7571 / 16695,
                393 / 640,
                -92097 / 339200,
                187 / 2100,
                1 / 40,
            ],
            dtype=self.options.ctype,
        )

        return alpha, beta, csol5, csol5 - csol4

    def step(
        self,
        t: float,
        y: Tensor,
        f: Tensor,
        dt: float,
        fun: Callable[[float, Tensor], Tensor],
    ) -> tuple[Tensor, Tensor, Tensor]:
        # import butcher tableau
        alpha, beta, csol, cerr = self.tableau

        # compute iterated Runge-Kutta values
        k = torch.empty(7, *f.shape, dtype=self.options.ctype)
        k[0] = f.to_dense() if self.options.use_sparse else f
        for i in range(1, 7):
            dy = torch.tensordot(dt * beta[i - 1, :i], k[:i].clone(), dims=([0], [0]))
            a = fun(t + dt * alpha[i - 1].item(), y + dy)
            k[i] = a.to_dense() if self.options.use_sparse else a

        # compute results
        f_new = k[-1]
        y_new = y + torch.tensordot(dt * csol[:6], k[:6], dims=([0], [0]))
        y_err = torch.tensordot(dt * cerr, k, dims=([0], [0]))

        return f_new, y_new, y_err



from __future__ import annotations

from functools import cached_property
from typing import Any, Callable

import torch
from torch import Tensor

from pyqtorch.embed import Embedding
from pyqtorch.matrices import OPERATIONS_DICT, controlled
from pyqtorch.noise import DigitalNoiseProtocol, _repr_noise
from pyqtorch.quantum_operation import QuantumOperation, Support
from pyqtorch.utils import DensityMatrix

from .mutation_utils import (
    mutate_control_slice,
    mutate_revert_modified,
    mutate_separate_target,
)


class Primitive(QuantumOperation):
    """Primitive operators based on a fixed matrix U.


    Attributes:
        operation (Tensor): Matrix U.
        qubit_support: List of qubits the QuantumOperation acts on.
        generator (Tensor): A tensor G s.t. U = exp(-iG).
        noise ( NoiseProtocol | dict[str, NoiseProtocol. optional): Type of noise
            to add in the operation.
        diagonal (bool, optional): Specify if the operation is diagonal.
    """

    def __init__(
        self,
        operation: Tensor,
        qubit_support: int | tuple[int, ...] | Support,
        generator: Tensor | None = None,
        noise: DigitalNoiseProtocol | None = None,
        diagonal: bool = False,
    ) -> None:
        super().__init__(operation, qubit_support, noise=noise, diagonal=diagonal)
        self.generator = generator

    def to(self, *args: Any, **kwargs: Any) -> Primitive:
        """Do device or dtype conversions.

        Returns:
            Primitive: Converted instance.
        """
        super().to(*args, **kwargs)
        if self.generator is not None:
            self.generator.to(*args, **kwargs)
        return self

    @cached_property
    def eigenvals_generator(self) -> Tensor:
        """Get eigenvalues of the underlying generator.

        Note that for a primitive, the generator is unclear
        so we execute pass.

        Arguments:
            values: Parameter values.

        Returns:
            Eigenvalues of the generator operator.
        """
        if self.generator is not None:
            return torch.linalg.eigvalsh(self.generator).reshape(-1, 1)
        pass

    @property
    def is_parametric(self) -> bool:
        return False


class ControlledPrimitive(Primitive):
    """Primitive applied depending on control qubits.

    Attributes:
        operation (Tensor): Unitary tensor U.
        control (int | tuple[int, ...]): List of qubits acting as controls.
        target (int | tuple[int, ...]): List of qubits operations acts on.
    """

    def __init__(
        self,
        operation: str | Tensor,
        control: int | tuple[int, ...],
        target: int | tuple[int, ...],
        noise: DigitalNoiseProtocol | None = None,
    ):
        support = Support(target, control)
        if isinstance(operation, str):
            operation = OPERATIONS_DICT[operation]
        operation = controlled(
            operation=operation.unsqueeze(2),
            batch_size=1,
            n_control_qubits=len(support.control),
        ).squeeze(2)
        super().__init__(operation, support, noise=noise)

    def extra_repr(self) -> str:
        return f"control: {self.control}, target: {self.target}" + _repr_noise(
            self.noise
        )


class MutablePrimitive(Primitive):
    """Primitive with a mutation operation via a callable `modifier`
    acting directly on the input state.

    Reference: https://arxiv.org/pdf/2303.01493

    Attributes:
        modifier (Callable): Function to modify the state, so applying a gate effect.
    """

    def __init__(
        self,
        operation: Tensor,
        target: int | tuple[int, ...],
        generator: Tensor | None = None,
        noise: DigitalNoiseProtocol | None = None,
        modifier: Callable = lambda s: s,
    ):
        super().__init__(operation, target, generator=generator, noise=noise)
        self.modifier: Callable = modifier

    def _mutate_state_vector(self, state: Tensor) -> Tensor:
        state_shape = state.shape
        perm, state = mutate_separate_target(state, self.target)

        state = self.modifier(state)

        return mutate_revert_modified(state, state_shape, perm)

    def _forward(
        self,
        state: Tensor,
        values: dict[str, Tensor] | Tensor | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        values = values or dict()
        if isinstance(state, DensityMatrix):
            return super()._forward(state, values, embedding)
        else:
            return self._mutate_state_vector(state)


class PhaseMutablePrimitive(MutablePrimitive):
    """Primitive with a mutation operation where we multiply by a phase given
    by the matrix element for the state 1.
    """

    def __init__(
        self,
        operation: Tensor,
        target: int,
        generator: Tensor | None = None,
        noise: DigitalNoiseProtocol | None = None,
    ):
        super().__init__(operation, target, generator=generator, noise=noise)
        self.modifier = self.apply_phase1

    def apply_phase1(self, state) -> Tensor:
        phase_mask = torch.ones_like(state, dtype=state.dtype, device=state.device)
        phase_mask[1, :] = self.operation[1, 1]
        phase_state = state * phase_mask
        return phase_state


class MutableControlledPrimitive(ControlledPrimitive):
    """Controlled primitive with a mutation operation via a callable `modifier`
    acting directly on the input state.

    Reference: https://arxiv.org/pdf/2303.01493

    Attributes:
        modifier (Callable): Function to modify the state, so applying a gate effect.
    """

    def __init__(
        self,
        operation: str | Tensor,
        control: int | tuple[int, ...],
        target: int | tuple[int, ...],
        noise: DigitalNoiseProtocol | None = None,
        modifier: Callable = lambda s: s,
    ):
        super().__init__(operation, control, target, noise=noise)
        self.modifier: Callable = modifier

    def _mutate_state_vector(self, state: Tensor) -> Tensor:
        result = state.clone()

        # first, create a mask to separate states to be modified
        mask = mutate_control_slice(len(state.shape) - 1, self.control)
        control_state = state[mask]
        # second, modify only the controlled states
        result[mask] = self._mutate_control_state(control_state)

        return result

    def _mutate_control_state(self, control_state: Tensor) -> Tensor:
        control_state_shape = control_state.shape
        perm, control_state = mutate_separate_target(control_state, self.target)

        control_state = self.modifier(control_state)
        control_state = mutate_revert_modified(control_state, control_state_shape, perm)
        return control_state

    def _forward(
        self,
        state: Tensor,
        values: dict[str, Tensor] | Tensor | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        values = values or dict()
        if isinstance(state, DensityMatrix):
            return super()._forward(state, values, embedding)
        else:
            return self._mutate_state_vector(state)



from __future__ import annotations

import logging
from functools import reduce
from logging import getLogger
from operator import add
from typing import Any, Iterator

import torch
from numpy import int64
from torch import Tensor, complex128, einsum
from torch import device as torch_device
from torch import dtype as torch_dtype
from torch.nn import Module, ModuleList, ParameterDict

from pyqtorch.embed import Embedding
from pyqtorch.matrices import _dagger, add_batch_dim
from pyqtorch.utils import (
    State,
)

logger = getLogger(__name__)


def forward_hook(*args, **kwargs) -> None:  # type: ignore[no-untyped-def]
    logger.debug("Forward complete")
    torch.cuda.nvtx.range_pop()


def pre_forward_hook(*args, **kwargs) -> None:  # type: ignore[no-untyped-def]
    logger.debug("Executing forward")
    torch.cuda.nvtx.range_push("QuantumCircuit.forward")


def backward_hook(*args, **kwargs) -> None:  # type: ignore[no-untyped-def]
    logger.debug("Backward complete")
    torch.cuda.nvtx.range_pop()


def pre_backward_hook(*args, **kwargs) -> None:  # type: ignore[no-untyped-def]
    logger.debug("Executed backward")
    torch.cuda.nvtx.range_push("QuantumCircuit.backward")


class Sequence(Module):
    """A generic container for pyqtorch operations"""

    def __init__(self, operations: list[Module]):
        super().__init__()
        self.operations = ModuleList(operations)
        self._device = torch_device("cpu")
        self._dtype = complex128
        if len(self.operations) > 0:
            try:
                self._device = next(iter(set((op.device for op in self.operations))))
            except StopIteration:
                pass
        logger.debug("QuantumCircuit initialized")
        if logger.isEnabledFor(logging.DEBUG):
            # When Debugging let's add logging and NVTX markers
            # WARNING: incurs performance penalty
            self.register_forward_hook(forward_hook, always_call=True)
            self.register_full_backward_hook(backward_hook)
            self.register_forward_pre_hook(pre_forward_hook)
            self.register_full_backward_pre_hook(pre_backward_hook)

        self._qubit_support = tuple(
            set(
                sum(
                    [op.qubit_support for op in self.operations],
                    (),
                )
            )
        )

        self._qubit_support = tuple(
            map(
                lambda x: x if isinstance(x, (int, int64)) else x[0],
                self._qubit_support,
            )
        )
        assert all(
            [isinstance(q, (int, int64)) for q in self._qubit_support]
        )  # TODO fix numpy.int issue

        self.is_diagonal = all([op.is_diagonal for op in self.flatten()])

    @property
    def qubit_support(self) -> tuple:
        return tuple(sorted(self._qubit_support))

    @property
    def params(self) -> list:
        return list(set(op.param_name for op in self.flatten() if op.is_parametric))

    @property
    def is_parametric(self) -> bool:
        return len(self.params) > 0

    def __iter__(self) -> Iterator:
        return iter(self.operations)

    def __len__(self) -> int:
        return len(self.operations)

    def __hash__(self) -> int:
        return hash(
            reduce(add, (hash(op) for op in self.operations))
            if len(self.operations) > 0
            else self.operations
        )

    def forward(
        self,
        state: Tensor,
        values: dict[str, Tensor] | ParameterDict | None = None,
        embedding: Embedding | None = None,
    ) -> State:
        values = values or dict()
        for op in self.operations:
            state = op(state, values, embedding)
        return state

    @property
    def device(self) -> torch_device:
        return self._device

    @property
    def dtype(self) -> torch_dtype:
        return self._dtype

    def to(self, *args: Any, **kwargs: Any) -> Sequence:
        self.operations = ModuleList([op.to(*args, **kwargs) for op in self.operations])
        if len(self.operations) > 0:
            self._device = self.operations[0].device
            self._dtype = self.operations[0].dtype
        return self

    def _flatten(self) -> ModuleList:
        """Method helper for checking commutation of pauli operators, especially for Scale."""
        ops = []
        for op in self.operations:
            if isinstance(op, Sequence):
                ops += op._flatten()
            else:
                ops.append(op)
        return ModuleList(ops)

    def flatten(self) -> ModuleList:
        ops = []
        for op in self.operations:
            if isinstance(op, Sequence):
                ops += op.flatten()
            else:
                ops.append(op)
        return ModuleList(ops)

    def tensor(
        self,
        values: dict[str, Tensor] | None = None,
        embedding: Embedding | None = None,
        full_support: tuple[int, ...] | None = None,
        diagonal: bool = False,
    ) -> Tensor:
        if full_support is None:
            full_support = self.qubit_support
        elif not set(self.qubit_support).issubset(set(full_support)):
            raise ValueError(
                "Expanding tensor operation requires a `full_support` argument "
                "larger than or equal to the `qubit_support`."
            )
        use_diagonal = diagonal and self.is_diagonal
        values = values or dict()
        if not use_diagonal:
            mat = torch.eye(
                2 ** len(full_support), dtype=self.dtype, device=self.device
            ).unsqueeze(2)

            return reduce(
                lambda t0, t1: einsum("ijb,jkb->ikb", t1, t0),
                (
                    add_batch_dim(op.tensor(values, embedding, full_support))
                    for op in self.operations
                ),
                mat,
            )
        else:
            mat = torch.ones(
                2 ** len(full_support), dtype=self.dtype, device=self.device
            ).unsqueeze(1)
            return reduce(
                lambda t0, t1: t0 * t1,
                (
                    op.tensor(values, embedding, full_support, diagonal=True)
                    for op in self.operations
                ),
                mat,
            )

    def dagger(
        self,
        values: dict[str, Tensor] | Tensor | None = None,
        embedding: Embedding | None = None,
        diagonal: bool = False,
    ) -> Tensor:
        values = values or dict()
        use_diagonal = diagonal and self.is_diagonal
        return _dagger(self.tensor(values, embedding, diagonal=use_diagonal))



from __future__ import annotations

import pytest
import torch

import pyqtorch as pyq
from pyqtorch.utils import ATOL, DropoutMode


@pytest.mark.parametrize(
    "dropout_mode",
    [DropoutMode.ROTATIONAL, DropoutMode.ENTANGLING, DropoutMode.CANONICAL_FWD],
)
def test_dropout(
    dropout_mode: DropoutMode, n_qubits: int = 4, theta: float = 0.6
) -> None:
    """Test when dropout prob = 1.0 if all gates the dropout method affects are dropped."""
    circ_ops_1 = [pyq.RY(i, f"theta_{str(i)}") for i in range(n_qubits)]
    circ_ops_2 = [pyq.CNOT(i, i + 1) for i in range(n_qubits - 1)]

    drop_circ_ops = circ_ops_1 + circ_ops_2

    operations = {
        DropoutMode.ROTATIONAL: circ_ops_2,
        DropoutMode.ENTANGLING: circ_ops_1,
        DropoutMode.CANONICAL_FWD: [],
    }

    values = {
        f"theta_{str(i)}": torch.tensor([theta], requires_grad=True)
        for i in range(n_qubits)
    }

    state = pyq.zero_state(n_qubits=n_qubits)
    circ = pyq.QuantumCircuit(n_qubits=n_qubits, operations=operations[dropout_mode])
    dropout_circ = pyq.DropoutQuantumCircuit(
        n_qubits=n_qubits,
        operations=drop_circ_ops,
        dropout_mode=dropout_mode,
        dropout_prob=1.0,
    )
    obs = pyq.Observable([pyq.Z(1)])

    exp_circ = pyq.expectation(circuit=circ, state=state, values=values, observable=obs)
    exp_dropout_circ = pyq.expectation(
        circuit=dropout_circ, state=state, values=values, observable=obs
    )

    assert torch.allclose(exp_circ, exp_dropout_circ, atol=ATOL)



from __future__ import annotations

import torch
from torch import Tensor

from pyqtorch.embed import Embedding
from pyqtorch.matrices import OPERATIONS_DICT
from pyqtorch.noise import DigitalNoiseProtocol
from pyqtorch.quantum_operation import Support
from pyqtorch.utils import (
    product_state,
    qubit_support_as_tuple,
)

from .mutation_utils import mutate_control_slice
from .primitive import (
    ControlledPrimitive,
    MutableControlledPrimitive,
    MutablePrimitive,
    PhaseMutablePrimitive,
    Primitive,
)


class X(MutablePrimitive):
    def __init__(
        self,
        target: int,
        noise: DigitalNoiseProtocol | None = None,
    ):
        super().__init__(
            OPERATIONS_DICT["X"],
            target,
            noise=noise,
        )

    def _mutate_state_vector(self, state: Tensor) -> Tensor:
        #  X gate flips the values at the specified dimension
        return torch.flip(state, [self.target[0]])


class Y(MutablePrimitive):
    def __init__(
        self,
        target: int,
        noise: DigitalNoiseProtocol | None = None,
    ):
        super().__init__(OPERATIONS_DICT["Y"], target, noise=noise)
        self.modifier = self._apply_phases

    def _apply_phases(self, state: Tensor) -> Tensor:
        state = torch.flip(state, dims=[0])
        state[0, :] *= self.operation[0, 1]
        state[1, :] *= self.operation[1, 0]

        return state


class Z(PhaseMutablePrimitive):
    def __init__(
        self,
        target: int,
        noise: DigitalNoiseProtocol | None = None,
    ):
        super().__init__(OPERATIONS_DICT["Z"], target, noise=noise)


class I(Primitive):  # noqa: E742
    def __init__(
        self,
        target: int,
        noise: DigitalNoiseProtocol | None = None,
    ):
        super().__init__(OPERATIONS_DICT["I"], target, noise=noise)

    def _forward(
        self,
        state: Tensor,
        values: dict[str, Tensor] | Tensor | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        return state


class H(Primitive):
    def __init__(
        self,
        target: int,
        noise: DigitalNoiseProtocol | None = None,
    ):
        super().__init__(OPERATIONS_DICT["H"], target, noise=noise)


class T(PhaseMutablePrimitive):
    def __init__(
        self,
        target: int,
        noise: DigitalNoiseProtocol | None = None,
    ):
        super().__init__(OPERATIONS_DICT["T"], target, noise=noise)


class S(PhaseMutablePrimitive):
    def __init__(
        self,
        target: int,
        noise: DigitalNoiseProtocol | None = None,
    ):
        super().__init__(
            OPERATIONS_DICT["S"],
            target,
            generator=0.5 * OPERATIONS_DICT["Z"],
            noise=noise,
        )


class SDagger(PhaseMutablePrimitive):
    def __init__(
        self,
        target: int,
        noise: DigitalNoiseProtocol | None = None,
    ):
        super().__init__(
            OPERATIONS_DICT["SDAGGER"],
            target,
            generator=-0.5 * OPERATIONS_DICT["Z"],
            noise=noise,
        )


class Projector(Primitive):
    def __init__(
        self,
        qubit_support: int | tuple[int, ...],
        ket: str,
        bra: str,
        noise: DigitalNoiseProtocol | None = None,
    ):

        qubit_support = qubit_support_as_tuple(qubit_support)
        if len(ket) != len(bra):
            raise ValueError("Input ket and bra bitstrings must be of same length.")
        if len(qubit_support) != len(ket):
            raise ValueError(
                "Qubit support must have the same number of qubits of ket and bra states."
            )
        ket_state = product_state(ket).flatten()
        bra_state = product_state(bra).flatten()
        super().__init__(
            OPERATIONS_DICT["PROJ"](ket_state, bra_state), qubit_support, noise=noise
        )


class N(MutablePrimitive):
    def __init__(
        self,
        target: int,
        noise: DigitalNoiseProtocol | None = None,
    ):
        super().__init__(OPERATIONS_DICT["N"], target, noise=noise)

        self.modifier = self._zero_out

    def _zero_out(self, state: Tensor) -> Tensor:
        cloned_state = state.clone()
        cloned_state[0, :] = self.operation[0][0]
        return cloned_state


class SWAP(MutablePrimitive):
    def __init__(
        self,
        i: int,
        j: int,
        noise: DigitalNoiseProtocol | None = None,
    ):
        super().__init__(OPERATIONS_DICT["SWAP"], (i, j), noise=noise)

    def _mutate_state_vector(self, state: Tensor) -> Tensor:
        #  SWAP gate flips the values at the specified dimensions
        return state.transpose(self.target[0], self.target[1])


class CSWAP(Primitive):
    def __init__(
        self,
        control: int,
        target: tuple[int, ...],
        noise: DigitalNoiseProtocol | None = None,
    ):
        if not isinstance(target, tuple) or len(target) != 2:
            raise ValueError("Target qubits must be a tuple with two qubits.")
        support = Support(target=qubit_support_as_tuple(control) + target)
        super().__init__(OPERATIONS_DICT["CSWAP"], support)


class CNOT(MutableControlledPrimitive):
    def __init__(
        self,
        control: int | tuple[int, ...],
        target: int,
        noise: DigitalNoiseProtocol | None = None,
    ):
        super().__init__("X", control, target, noise=noise)

    def _mutate_control_state(self, control_state: Tensor) -> Tensor:
        target_adjusted = self.target[0] - sum(
            (1 for c in self.control if c < self.target[0])
        )
        return torch.flip(control_state, [target_adjusted])


CX = CNOT


class CY(ControlledPrimitive):
    def __init__(
        self,
        control: int | tuple[int, ...],
        target: int,
        noise: DigitalNoiseProtocol | None = None,
    ):
        super().__init__("Y", control, target, noise=noise)


class CZ(MutableControlledPrimitive):
    def __init__(
        self,
        control: int | tuple[int, ...],
        target: int,
        noise: DigitalNoiseProtocol | None = None,
    ):
        super().__init__("Z", control, target, noise=noise)

    def _mutate_state_vector(self, state):
        result = state.clone()
        all_qubits = self.control + self.target
        phase_selector = mutate_control_slice(len(state.shape) - 1, all_qubits)
        result[phase_selector] = -result[phase_selector]
        return result


class Toffoli(MutableControlledPrimitive):
    def __init__(
        self,
        control: int | tuple[int, ...],
        target: int,
        noise: DigitalNoiseProtocol | None = None,
    ):
        super().__init__("X", control, target, noise=noise)

    def _mutate_control_state(self, control_state: Tensor) -> Tensor:
        target_adjusted = self.target[0] - sum(
            (1 for c in self.control if c < self.target[0])
        )
        return torch.flip(control_state, [target_adjusted])


OPS_PAULI = {X, Y, Z, I}
OPS_1Q = OPS_PAULI.union({H, S, T})
OPS_2Q = {CNOT, CY, CZ, SWAP}
OPS_3Q = {Toffoli, CSWAP}
OPS_DIGITAL = OPS_1Q.union(OPS_2Q, OPS_3Q)

OPS_DIAGONAL = {Z, I, S, T, SDagger, N, CZ}



from __future__ import annotations

import random

import pytest
import torch
from helpers import get_op_support, random_pauli_hamiltonian
from torch import Tensor

from pyqtorch.apply import apply_operator_dm, operator_product
from pyqtorch.circuit import QuantumCircuit
from pyqtorch.hamiltonians import Observable
from pyqtorch.matrices import (
    HMAT,
    IMAT,
    XMAT,
    YMAT,
    ZMAT,
    _dagger,
)
from pyqtorch.noise import (
    AmplitudeDamping,
    AnalogDepolarizing,
    DigitalNoiseProtocol,
    DigitalNoiseType,
    GeneralizedAmplitudeDamping,
    Noise,
    PhaseDamping,
)
from pyqtorch.primitives import (
    OPS_DIGITAL,
    OPS_PARAM,
    ControlledPrimitive,
    ControlledRotationGate,
    H,
    I,
    Parametric,
    Primitive,
    X,
    Y,
    Z,
)
from pyqtorch.utils import (
    ATOL,
    RTOL,
    DensityMatrix,
    density_mat,
    operator_kron,
    product_state,
    random_state,
)


def test_dm(n_qubits: int, batch_size: int) -> None:
    state = random_state(n_qubits)
    projector = torch.outer(state.flatten(), state.conj().flatten()).view(
        2**n_qubits, 2**n_qubits, 1
    )
    dm = density_mat(state)
    assert dm.size() == torch.Size([2**n_qubits, 2**n_qubits, 1])
    assert torch.allclose(dm, projector)
    assert torch.allclose(dm.squeeze(), dm.squeeze() @ dm.squeeze())
    states = []
    projectors = []
    for batch in range(batch_size):
        state = random_state(n_qubits)
        states.append(state)
        projector = torch.outer(state.flatten(), state.conj().flatten()).view(
            2**n_qubits, 2**n_qubits, 1
        )
        projectors.append(projector)
    dm_proj = torch.cat(projectors, dim=2)
    state_cat = torch.cat(states, dim=n_qubits)
    dm = density_mat(state_cat)
    assert dm.size() == torch.Size([2**n_qubits, 2**n_qubits, batch_size])
    assert torch.allclose(dm, dm_proj)


@pytest.mark.parametrize("n_qubits", [{"low": 2, "high": 5}], indirect=True)
def test_operator_product(
    random_unitary_gate: Primitive | Parametric,
    n_qubits: int,
) -> None:
    batch_size = torch.randint(low=2, high=5, size=(1,)).item()
    values = {"theta": torch.rand(1)}
    full_support = tuple(range(n_qubits))
    op = random_unitary_gate.tensor(values=values, full_support=full_support)
    op_mul = operator_product(
        op1=op.repeat(1, 1, batch_size),
        supp1=full_support,
        op2=_dagger(op),
        supp2=full_support,
    )
    assert op_mul.size() == torch.Size([2**n_qubits, 2**n_qubits, batch_size])
    assert torch.allclose(
        op_mul,
        torch.eye(2**n_qubits, dtype=torch.cdouble)
        .unsqueeze(2)
        .repeat(1, 1, batch_size),
    )


@pytest.mark.parametrize("n_qubits", [{"low": 2, "high": 5}], indirect=True)
def test_apply_density_mat(
    random_unitary_gate: Primitive | Parametric,
    n_qubits: int,
    batch_size: int,
    random_input_dm: DensityMatrix,
) -> None:
    values = {"theta": torch.rand(1)}
    full_support = tuple(range(n_qubits))
    op = random_unitary_gate
    rho = random_input_dm
    op_mat = op.tensor(values=values)
    rho_evol = apply_operator_dm(rho, op_mat, op.qubit_support)
    assert rho_evol.size() == torch.Size([2**n_qubits, 2**n_qubits, batch_size])
    mul1 = operator_product(rho, full_support, _dagger(op_mat), op.qubit_support)
    rho_expected = operator_product(op_mat, op.qubit_support, mul1, full_support)
    assert torch.allclose(rho_evol, rho_expected)


@pytest.mark.parametrize(
    "operator,matrix", [(I, IMAT), (X, XMAT), (Z, ZMAT), (Y, YMAT), (H, HMAT)]
)
def test_operator_kron(operator: Tensor, matrix: Tensor) -> None:
    n_qubits = torch.randint(low=1, high=5, size=(1,)).item()
    batch_size = torch.randint(low=1, high=5, size=(1,)).item()
    states, krons = [], []
    for batch in range(batch_size):
        state = random_state(n_qubits)
        states.append(state)
        kron = torch.kron(density_mat(state).squeeze(), matrix).unsqueeze(2)
        krons.append(kron)
    input_state = torch.cat(states, dim=n_qubits)
    kron_out = operator_kron(density_mat(input_state), operator(0).dagger())
    assert kron_out.size() == torch.Size(
        [2 ** (n_qubits + 1), 2 ** (n_qubits + 1), batch_size]
    )
    kron_expect = torch.cat(krons, dim=2)
    assert torch.allclose(kron_out, kron_expect)
    assert torch.allclose(
        torch.kron(operator(0).dagger().contiguous(), I(0).tensor()),
        operator_kron(operator(0).dagger(), I(0).tensor()),
    )


def test_kron_batch() -> None:
    n_qubits = torch.randint(low=1, high=5, size=(1,)).item()
    batch_size_1 = torch.randint(low=1, high=5, size=(1,)).item()
    batch_size_2 = torch.randint(low=1, high=5, size=(1,)).item()
    max_batch = max(batch_size_1, batch_size_2)
    dm_1 = density_mat(random_state(n_qubits, batch_size_1))
    dm_2 = density_mat(random_state(n_qubits, batch_size_2))
    dm_out = operator_kron(dm_1, dm_2)
    assert dm_out.size() == torch.Size(
        [2 ** (2 * n_qubits), 2 ** (2 * n_qubits), max_batch]
    )
    if batch_size_1 > batch_size_2:
        dm_2 = dm_2.repeat(1, 1, batch_size_1)[:, :, :batch_size_1]
    elif batch_size_2 > batch_size_1:
        dm_1 = dm_1.repeat(1, 1, batch_size_2)[:, :, :batch_size_2]
    density_matrices = []
    for batch in range(max_batch):
        density_matrice = torch.kron(dm_1[:, :, batch], dm_2[:, :, batch]).unsqueeze(2)
        density_matrices.append(density_matrice)
    dm_expect = torch.cat(density_matrices, dim=2)
    assert torch.allclose(dm_out, dm_expect)


@pytest.mark.parametrize("n_qubits", [{"low": 2, "high": 5}], indirect=True)
def test_flip_gates(
    n_qubits: int,
    target: int,
    batch_size: int,
    rho_input: Tensor,
    random_flip_gate: Noise,
    flip_expected_state: DensityMatrix,
    flip_probability: Tensor | float,
    flip_gates_prob_0: Noise,
    flip_gates_prob_1: tuple,
    random_input_dm: DensityMatrix,
) -> None:
    FlipGate = random_flip_gate
    output_state: DensityMatrix = FlipGate(target, flip_probability)(rho_input)
    assert output_state.size() == torch.Size([2**n_qubits, 2**n_qubits, batch_size])
    assert torch.allclose(output_state, flip_expected_state)

    input_state = random_input_dm  # fix the same random state for every call
    assert torch.allclose(flip_gates_prob_0(input_state), input_state)

    FlipGate_1, expected_op = flip_gates_prob_1
    assert torch.allclose(FlipGate_1(input_state), expected_op)


def test_damping_gates(
    n_qubits: int,
    target: int,
    batch_size: int,
    damping_expected_state: tuple,
    damping_gates_prob_0: Tensor,
    random_input_dm: DensityMatrix,
    rho_input: Tensor,
) -> None:
    DampingGate, expected_state = damping_expected_state
    apply_gate = DampingGate(rho_input)
    assert apply_gate.size() == torch.Size([2**n_qubits, 2**n_qubits, batch_size])
    assert torch.allclose(apply_gate, expected_state)

    input_state = random_input_dm
    assert torch.allclose(
        damping_gates_prob_0(input_state),
        I(target)(input_state),
    )

    rho_0: DensityMatrix = density_mat(product_state("0", batch_size))
    rho_1: DensityMatrix = density_mat(product_state("1", batch_size))
    if DampingGate == AmplitudeDamping:
        assert torch.allclose(DampingGate(target, rate=1)(rho_1), rho_0)
    elif DampingGate == PhaseDamping:
        assert torch.allclose(DampingGate(target, rate=1)(rho_1), I(target)(rho_1))
    elif DampingGate == GeneralizedAmplitudeDamping:
        assert torch.allclose(
            DampingGate(target, error_probability=(1, 1))(rho_1), rho_0
        )


@pytest.mark.parametrize("n_qubits", [{"low": 2, "high": 5}], indirect=True)
def test_noisy_primitive(
    random_noisy_unitary_gate: tuple,
    random_input_dm: DensityMatrix,
    n_qubits: int,
    batch_size: int,
) -> None:
    noisy_primitive, primitve_gate, noise_gate = random_noisy_unitary_gate
    state = random_input_dm
    values = {"theta": torch.rand(1)}
    rho_evol = noisy_primitive(state, values)
    assert rho_evol.size() == torch.Size([2**n_qubits, 2**n_qubits, batch_size])
    rho_expected = noise_gate(primitve_gate(state, values))
    assert rho_expected.size() == torch.Size([2**n_qubits, 2**n_qubits, batch_size])
    assert torch.allclose(rho_evol, rho_expected)


@pytest.mark.parametrize("n_qubits", [{"low": 2, "high": 5}], indirect=True)
@pytest.mark.parametrize("batch_size", [{"low": 1, "high": 5}], indirect=True)
def test_noise_circ(
    n_qubits: int,
    batch_size: int,
    random_input_dm: DensityMatrix,
    random_single_qubit_gate: Primitive,
    random_noise_gate: Noise,
    random_rotation_gate: Parametric,
    random_controlled_gate: ControlledPrimitive,
    random_rotation_control_gate: ControlledRotationGate,
) -> None:
    OPERATORS = [
        random_single_qubit_gate,
        random_noise_gate,
        random_rotation_gate,
        random_controlled_gate,
        random_rotation_control_gate,
    ]
    random.shuffle(OPERATORS)
    circ = QuantumCircuit(n_qubits, OPERATORS)

    values = {random_rotation_gate.param_name: torch.rand(1)}
    output_state = circ(random_input_dm, values)
    assert isinstance(output_state, DensityMatrix)
    assert output_state.shape == torch.Size([2**n_qubits, 2**n_qubits, batch_size])

    diag_sums = []
    for i in range(batch_size):
        diag_batch = torch.diagonal(output_state[:, :, i], dim1=0, dim2=1)
        diag_sums.append(torch.sum(diag_batch))
    diag_sum = torch.stack(diag_sums)
    assert torch.allclose(diag_sum, torch.ones((batch_size,), dtype=torch.cdouble))


@pytest.mark.parametrize("make_param", [True, False])
@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("commuting_terms", [False, True])
def test_dm_expectation(
    n_qubits: int, batch_size: int, make_param: bool, commuting_terms: bool
) -> None:
    k_1q = 2 * n_qubits  # Number of 1-qubit terms
    k_2q = n_qubits**2  # Number of 2-qubit terms
    hamiltonian, param_list = random_pauli_hamiltonian(
        n_qubits, k_1q, k_2q, make_param, commuting_terms=commuting_terms
    )
    values = {param: torch.rand(batch_size) for param in param_list}
    psi_init = random_state(n_qubits, batch_size)
    dm_init = density_mat(psi_init)
    obs = Observable(hamiltonian)
    exp_state = obs.expectation(psi_init, values=values)
    exp_dm = obs.expectation(dm_init, values=values)
    assert torch.allclose(exp_state, exp_dm)


@pytest.mark.parametrize("noise_type", [noise for noise in DigitalNoiseType])
@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("batch_size", [1, 5])
def test_digital_noise_apply(
    n_qubits: int,
    batch_size: int,
    noise_type: DigitalNoiseType,
) -> None:
    """
    Goes through all non-parametric gates and tests their application to a random state
    in comparison with the noisy version with error_probability = 0.
    """
    op: type[Primitive]
    error_probability: float | tuple[float, ...]

    if noise_type == DigitalNoiseType.PAULI_CHANNEL:
        error_probability = (0.0, 0.0, 0.0)
    elif noise_type == DigitalNoiseType.GENERALIZED_AMPLITUDE_DAMPING:
        error_probability = (0.0, 0.0)
    else:
        error_probability = 0.0

    noise_concrete = DigitalNoiseProtocol(noise_type, error_probability)

    for op in OPS_DIGITAL:
        supp = get_op_support(op, n_qubits)
        op_concrete = op(*supp)
        op_concrete_noise = op(*supp, noise=noise_concrete)  # type: ignore [misc]
        psi_init = density_mat(random_state(n_qubits, batch_size))
        psi_expected = op_concrete(psi_init)
        psi_star = op_concrete_noise(psi_init)
        assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("noise_type", [noise for noise in DigitalNoiseType])
@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("batch_size", [1, 5])
def test_param_noise_apply(
    n_qubits: int,
    batch_size: int,
    noise_type: DigitalNoiseType,
) -> None:
    """
    Goes through all parametric gates and tests their application to a random state
    in comparison with the noisy version with error_probability = 0.
    """
    op: type[Parametric]

    error_probability: float | tuple[float, ...]

    if noise_type == DigitalNoiseType.PAULI_CHANNEL:
        error_probability = (0.0, 0.0, 0.0)
    elif noise_type == DigitalNoiseType.GENERALIZED_AMPLITUDE_DAMPING:
        error_probability = (0.0, 0.0)
    else:
        error_probability = 0.0

    noise_concrete = DigitalNoiseProtocol(noise_type, error_probability)

    for op in OPS_PARAM:
        supp = get_op_support(op, n_qubits)
        params = [f"th{i}" for i in range(op.n_params)]
        op_concrete = op(*supp, *params)
        op_concrete_noise = op(*supp, *params, noise=noise_concrete)  # type: ignore [misc]
        psi_init = density_mat(random_state(n_qubits))
        values = {param: torch.rand(batch_size) for param in params}
        psi_expected = op_concrete(psi_init, values=values)
        psi_star = op_concrete_noise(psi_init, values=values)
        assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)


def test_analog_noise_add():
    noise1 = AnalogDepolarizing(error_param=0.1, qubit_support=3)
    noise2 = AnalogDepolarizing(error_param=0.2, qubit_support=3)
    noise3 = AnalogDepolarizing(error_param=0.2, qubit_support=2)

    noise_add = noise1 + noise2
    assert len(noise1.noise_operators) + len(noise2.noise_operators) == len(
        noise_add.noise_operators
    )
    assert noise1.qubit_support == noise_add.qubit_support

    noise_add = noise1 + noise3
    assert len(noise1.noise_operators) + len(noise3.noise_operators) == len(
        noise_add.noise_operators
    )
    assert noise_add.qubit_support == (2, 3)



from __future__ import annotations

import random

import numpy as np
import torch

import pyqtorch.embed as pyq_em
from pyqtorch.apply import apply_operator, apply_operator_permute
from pyqtorch.composite import Add, Scale, Sequence
from pyqtorch.hamiltonians.evolution import HamiltonianEvolution
from pyqtorch.primitives import (
    OPS_1Q,
    OPS_2Q,
    OPS_3Q,
    OPS_PARAM_1Q,
    OPS_PARAM_2Q,
    OPS_PAULI,
    I,
    Parametric,
    Primitive,
    Toffoli,
    Z,
)


def calc_mat_vec_wavefunction(
    block: Primitive | Sequence,
    init_state: torch.Tensor,
    values: dict | None = None,
    full_support: tuple | None = None,
    use_permute: bool = False,
) -> torch.Tensor:
    """Get the result of applying the matrix representation of a block to an initial state.

    Args:
        block: The black operator to apply.
        init_state: Initial state to apply block on.
        values: Values of parameter if block is parametric.

    Returns:
        Tensor: The new wavefunction after applying the block.

    """
    use_diagonal = (
        block.is_diagonal if not isinstance(block, HamiltonianEvolution) else False
    )
    mat = block.tensor(values=values, full_support=full_support, diagonal=use_diagonal)
    qubit_support = block.qubit_support if full_support is None else full_support
    apply_func = apply_operator_permute if use_permute else apply_operator
    return apply_func(
        init_state,
        mat,
        qubit_support,
        diagonal=use_diagonal,
    )


def get_op_support(
    op: type[Primitive] | type[Parametric], n_qubits: int, get_ordered: bool = False
) -> tuple:
    """Decides a random qubit support for any gate, up to a some max n_qubits."""
    if op in OPS_1Q.union(OPS_PARAM_1Q):
        supp: tuple = (random.randint(0, n_qubits - 1),)
        ordered_supp = supp
    elif op in OPS_2Q.union(OPS_PARAM_2Q):
        supp = tuple(random.sample(range(n_qubits), 2))
        ordered_supp = tuple(sorted(supp))
    elif op in OPS_3Q:
        i, j, k = tuple(random.sample(range(n_qubits), 3))
        a, b, c = tuple(sorted((i, j, k)))
        supp = ((i, j), k) if op == Toffoli else (i, (j, k))
        ordered_supp = ((a, b), c) if op == Toffoli else (a, (b, c))
    if get_ordered:
        return supp, ordered_supp
    else:
        return supp


def get_random_embed() -> tuple:
    fn_list = [
        (pyq_em.sin, torch.sin),
        (pyq_em.cos, torch.cos),
        (pyq_em.log, torch.log),
        (pyq_em.tanh, torch.tanh),
        (pyq_em.tan, torch.tan),
        (pyq_em.sqrt, torch.sqrt),
    ]

    fn1, fn2 = random.choice(fn_list), random.choice(fn_list)

    expr = (1.0 + 2 ** fn1[0]("x")) * fn2[0]("x")
    call = lambda x: (12.0 + 2 ** fn1[1](x)) * fn2[1]("x")

    embedding = pyq_em.Embedding(fparam_names=["x"], var_to_call={"expr": expr})

    return embedding, call


def random_pauli_hamiltonian(
    n_qubits: int,
    k_1q: int = 5,
    k_2q: int = 10,
    make_param: bool = False,
    default_scale_coeffs: float | None = None,
    p_param: float = 0.5,
    diagonal: bool = False,
    commuting_terms: bool = False,
) -> tuple[Sequence, list]:
    """Creates a random Pauli Hamiltonian as a sum of k_1q + k_2q terms.

    Args:
        n_qubits (int): Number of qubits.
        k_1q (int, optional): Number of one-qubit terms. Defaults to 5.
        k_2q (int, optional): Number of two-qubit terms. Defaults to 10.
        make_param (bool, optional): Coefficients as parameters. Defaults to False.
        default_scale_coeffs (float | None, optional): Default value for the parameter
            of Scale operations. Defaults to None.
        p_param (float): Probability for each term to be chosen to add a parameter.

    Returns:
        tuple[Sequence, list]: Hamiltonian and list of parameters.
    """
    OPS_PAULI_choice = list(OPS_PAULI) if not diagonal else [I, Z]
    if commuting_terms:
        k_1q = n_qubits
        k_2q = 0
    one_q_terms: list = random.choices(OPS_PAULI_choice, k=k_1q)

    two_q_terms: list = [random.choices(OPS_PAULI_choice, k=2) for _ in range(k_2q)]
    terms: list = []
    for i, term in enumerate(one_q_terms):
        if commuting_terms:
            supp = [i]
        else:
            supp = random.sample(range(n_qubits), 1)
        terms.append(term(supp))
    for term in two_q_terms:
        supp = random.sample(range(n_qubits), 2)
        terms.append(Sequence([term[0](supp[0]), term[1](supp[1])]))
    param_list = []
    for i, t in enumerate(terms):
        if random.random() < p_param:
            if make_param:
                terms[i] = Scale(t, f"p_{i}")
                param_list.append(f"p_{i}")
            else:
                terms[i] = Scale(
                    t,
                    (
                        torch.rand(1)
                        if not default_scale_coeffs
                        else torch.tensor([default_scale_coeffs])
                    ),
                )
    return Add(terms), param_list


def random_parameter_names(names: list[str]) -> list[str]:
    return np.random.choice(names, len(names), replace=True)



from __future__ import annotations

from functools import cached_property
from typing import Any

import torch
from torch import Tensor

from pyqtorch.embed import Embedding
from pyqtorch.matrices import (
    DEFAULT_MATRIX_DTYPE,
    controlled,
)
from pyqtorch.noise import DigitalNoiseProtocol

from .parametric import ControlledRotationGate, Parametric

pauli_singleq_eigenvalues = torch.tensor([[-1.0], [1.0]], dtype=torch.cdouble)


class RX(Parametric):
    """
    Primitive for the RX gate.

    The corresponding unitary is obtained by :math:`exp(-i X, t)`
    where :math:`t` is an input parameter
    and :math:`X` is the matrix representation of the X gate.

    Attributes:
        param_name: Name of parameters.
        parse_values: Method defining how to handle the values dictionary input.
    """

    def __init__(
        self,
        target: int,
        param_name: str | int | float | torch.Tensor = "",
        noise: DigitalNoiseProtocol | None = None,
    ):
        """Initializes RX.

        Arguments:
            target: Target qubit.
            param_name: Name of parameters.
            noise: Optional noise protocols to apply.
        """
        super().__init__("X", target, param_name, noise=noise)

    @cached_property
    def eigenvals_generator(self) -> Tensor:
        """Get eigenvalues of the underlying generator.

        Arguments:
            values: Parameter values.

        Returns:
            Eigenvalues of the generator operator.
        """
        return pauli_singleq_eigenvalues.to(device=self.device, dtype=self.dtype)


class RY(Parametric):
    """
    Primitive for the RY gate.

    The corresponding unitary is obtained by :math:`exp(-i Y, t)`
    where :math:`t` is an input parameter
    and :math:`Y` is the matrix representation of the Y gate.

    Attributes:
        param_name: Name of parameters.
        parse_values: Method defining how to handle the values dictionary input.
    """

    def __init__(
        self,
        target: int,
        param_name: str | int | float | torch.Tensor = "",
        noise: DigitalNoiseProtocol | None = None,
    ):
        """Initializes RY.

        Arguments:
            target: Target qubit.
            param_name: Name of parameters.
            noise: Optional noise protocols to apply.
        """
        super().__init__("Y", target, param_name, noise=noise)

    @cached_property
    def eigenvals_generator(self) -> Tensor:
        """Get eigenvalues of the underlying generator.

        Arguments:
            values: Parameter values.

        Returns:
            Eigenvalues of the generator operator.
        """
        return pauli_singleq_eigenvalues.to(device=self.device, dtype=self.dtype)


class RZ(Parametric):
    """
    Primitive for the RZ gate.

    The corresponding unitary is obtained by :math:`exp(-i Z, t)`
    where :math:`t` is an input parameter
    and :math:`Z` is the matrix representation of the Z gate.

    Attributes:
        param_name: Name of parameters.
        parse_values: Method defining how to handle the values dictionary input.
    """

    def __init__(
        self,
        target: int,
        param_name: str | int | float | torch.Tensor = "",
        noise: DigitalNoiseProtocol | None = None,
    ):
        """Initializes RZ.

        Arguments:
            target: Target qubit.
            param_name: Name of parameters.
            noise: Optional noise protocols to apply.
        """
        super().__init__("Z", target, param_name, noise=noise)

    @cached_property
    def eigenvals_generator(self) -> Tensor:
        """Get eigenvalues of the underlying generator.

        Arguments:
            values: Parameter values.

        Returns:
            Eigenvalues of the generator operator.
        """
        return pauli_singleq_eigenvalues.to(device=self.device, dtype=self.dtype)


class PHASE(Parametric):
    """
    Primitive for the PHASE gate.

    The corresponding unitary is obtained by :math:`exp(-i I, t)`
    where :math:`t` is an input parameter
    and :math:`I` is the identity.

    Attributes:
        param_name: Name of parameters.
        parse_values: Method defining how to handle the values dictionary input.
    """

    def __init__(
        self,
        target: int,
        param_name: str | int | float | torch.Tensor = "",
        noise: DigitalNoiseProtocol | None = None,
    ):
        """Initializes PHASE.

        Arguments:
            target: Target qubit.
            param_name: Name of parameters.
            noise: Optional noise protocols to apply.
        """
        super().__init__("I", target, param_name, noise=noise)

    @cached_property
    def eigenvals_generator(self) -> Tensor:
        """Get eigenvalues of the underlying generator.

        Arguments:
            values: Parameter values.

        Returns:
            Eigenvalues of the generator operator.
        """
        return torch.tensor([[0.0], [2.0]], dtype=self.dtype, device=self.device)

    def _construct_parametric_base_op(
        self,
        values: dict[str, Tensor] | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        """
        Get the corresponding unitary.

        Arguments:
            values: A dict containing a Parameter name and value.
            embedding: An optional embedding for parameters.

        Returns:
            The unitary representation.
        """
        values = values or dict()
        thetas = self.parse_values(values, embedding)
        batch_size = len(thetas)
        batch_mat = self.identity.unsqueeze(2).repeat(1, 1, batch_size)
        batch_mat[1, 1, :] = torch.exp(1.0j * thetas).unsqueeze(0).unsqueeze(1)
        return batch_mat

    def jacobian(
        self,
        values: dict[str, Tensor] | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        """
        Get the corresponding unitary of the jacobian.

        Arguments:
            values: Parameter value.

        Returns:
            The unitary representation of the jacobian.
        """
        values = values or dict()
        thetas = self.parse_values(values, embedding)
        batch_mat = (
            torch.zeros((2, 2), dtype=self.identity.dtype)
            .unsqueeze(2)
            .repeat(1, 1, len(thetas))
        )
        batch_mat[1, 1, :] = 1j * torch.exp(1j * thetas).unsqueeze(0).unsqueeze(1)
        return batch_mat


class CRX(ControlledRotationGate):
    """
    Primitive for the controlled RX gate.
    """

    def __init__(
        self,
        control: int | tuple[int, ...],
        target: int,
        param_name: str | int | float | Tensor = "",
        noise: DigitalNoiseProtocol | None = None,
    ):
        """Initializes controlled RX.

        Arguments:
            control: Control qubit(s).
            target: Target qubit.
            param_name: Name of parameters.
            noise: Optional noise protocols to apply.
        """
        super().__init__("X", control, target, param_name, noise=noise)

    @cached_property
    def eigenvals_generator(self) -> Tensor:
        """Get eigenvalues of the underlying generator.

        Arguments:
            values: Parameter values.

        Returns:
            Eigenvalues of the generator operator.
        """
        return torch.cat(
            (
                torch.zeros(
                    2 ** len(self.qubit_support) - 2,
                    device=self.device,
                    dtype=self.dtype,
                ),
                pauli_singleq_eigenvalues.flatten().to(
                    device=self.device, dtype=self.dtype
                ),
            )
        ).reshape(-1, 1)


class CRY(ControlledRotationGate):
    """
    Primitive for the controlled RY gate.
    """

    def __init__(
        self,
        control: int | tuple[int, ...],
        target: int,
        param_name: str | int | float | Tensor = "",
        noise: DigitalNoiseProtocol | None = None,
    ):
        """Initializes controlled RY.

        Arguments:
            control: Control qubit(s).
            target: Target qubit.
            param_name: Name of parameters.
            noise: Optional noise protocols to apply.
        """
        super().__init__("Y", control, target, param_name, noise=noise)

    @cached_property
    def eigenvals_generator(self) -> Tensor:
        """Get eigenvalues of the underlying generator.

        Arguments:
            values: Parameter values.

        Returns:
            Eigenvalues of the generator operator.
        """
        return torch.cat(
            (
                torch.zeros(
                    2 ** len(self.qubit_support) - 2,
                    device=self.device,
                    dtype=self.dtype,
                ),
                pauli_singleq_eigenvalues.flatten().to(
                    device=self.device, dtype=self.dtype
                ),
            )
        ).reshape(-1, 1)


class CRZ(ControlledRotationGate):
    """
    Primitive for the controlled RZ gate.
    """

    def __init__(
        self,
        control: int | tuple[int, ...],
        target: int,
        param_name: str | int | float | Tensor = "",
        noise: DigitalNoiseProtocol | None = None,
    ):
        """Initializes controlled RZ.

        Arguments:
            control: Control qubit(s).
            target: Target qubit.
            param_name: Name of parameters.
            noise: Optional noise protocols to apply.
        """
        super().__init__("Z", control, target, param_name, noise=noise)

    @cached_property
    def eigenvals_generator(self) -> Tensor:
        """Get eigenvalues of the underlying generator.

        Arguments:
            values: Parameter values.

        Returns:
            Eigenvalues of the generator operator.
        """
        return torch.cat(
            (
                torch.zeros(
                    2 ** len(self.qubit_support) - 2,
                    device=self.device,
                    dtype=self.dtype,
                ),
                pauli_singleq_eigenvalues.flatten().to(
                    device=self.device, dtype=self.dtype
                ),
            )
        ).reshape(-1, 1)


class CPHASE(ControlledRotationGate):
    """
    Primitive for the controlled PHASE gate.
    """

    n_params = 1

    def __init__(
        self,
        control: int | tuple[int, ...],
        target: int,
        param_name: str | int | float | Tensor = "",
        noise: DigitalNoiseProtocol | None = None,
    ):
        """Initializes controlled PHASE.

        Arguments:
            control: Control qubit(s).
            target: Target qubit.
            param_name: Name of parameters.
            noise: Optional noise protocols to apply.
        """
        super().__init__("I", control, target, param_name, noise=noise)

    @cached_property
    def eigenvals_generator(self) -> Tensor:
        """Get eigenvalues of the underlying generator.

        Arguments:
            values: Parameter values.

        Returns:
            Eigenvalues of the generator operator.
        """
        return torch.cat(
            (
                torch.tensor([-2.0, 0.0], device=self.device, dtype=self.dtype),
                torch.zeros(
                    2 ** len(self.qubit_support) - 2,
                    device=self.device,
                    dtype=self.dtype,
                ),
            )
        ).reshape(-1, 1)

    def _construct_parametric_base_op(
        self,
        values: dict[str, Tensor] | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        """
        Get the corresponding unitary.

        Arguments:
            values: A dict containing a Parameter name and value.
            embedding: An optional embedding for parameters.

        Returns:
            The unitary representation.
        """
        values = values or dict()
        thetas = self.parse_values(values, embedding)
        batch_size = len(thetas)
        mat = self.identity.unsqueeze(2).repeat(1, 1, batch_size)
        mat[1, 1, :] = torch.exp(1.0j * thetas).unsqueeze(0).unsqueeze(1)
        return controlled(mat, batch_size, len(self.control))

    def jacobian(
        self,
        values: dict[str, Tensor] | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        """
        Get the corresponding unitary of the jacobian.

        Arguments:
            values: Parameter value.

        Returns:
            The unitary representation of the jacobian.
        """
        values = values or dict()
        thetas = self.parse_values(values, embedding)
        batch_size = len(thetas)
        n_control = len(self.control)
        jU = (
            torch.zeros((2, 2), dtype=self.identity.dtype)
            .unsqueeze(2)
            .repeat(1, 1, len(thetas))
        )
        jU[1, 1, :] = 1j * torch.exp(1j * thetas).unsqueeze(0).unsqueeze(1)
        n_dim = 2 ** (n_control + 1)
        jC = (
            torch.zeros((n_dim, n_dim), dtype=self.identity.dtype)
            .unsqueeze(2)
            .repeat(1, 1, batch_size)
        )
        unitary_idx = 2 ** (n_control + 1) - 2
        jC[unitary_idx:, unitary_idx:, :] = jU
        return jC

    def to(self, *args: Any, **kwargs: Any) -> Parametric:
        """Set device of primitive.

        Returns:
            Primitive with device.
        """
        super().to(*args, **kwargs)
        self._device = self.identity.device
        return self


class U(Parametric):
    r"""
    Primitive for the U gate.

    The corresponding unitary representation is:
    .. math::

        U = \begin{bmatrix}
                cos(\frac{\theta}{2}) & 4  \\ e^{-i \omega} sin(\frac{\theta}{2})
                e^{i \phi} sin(\frac{\theta}{2}) & e^{i (\phi+ \omega)} cos(\frac{\theta}{2})
            \end{bmatrix}

    where :math:`\phi`, :math:`\omega` and :math:`\theta` are input parameters.
    Attributes:
        phi: Phi parameter.
        theta: Theta parameter.
        omega: Omega parameter.
    """

    n_params = 3

    def __init__(
        self,
        target: int,
        phi: str,
        theta: str,
        omega: str,
        noise: DigitalNoiseProtocol | None = None,
    ):
        """Initializes U gate.

        Arguments:
            target: Target qubit.
            phi: Phi parameter.
            theta: Theta parameter.
            omega: Omega parameter.
            noise: Optional noise protocols to apply.
        """
        self.phi = phi
        self.theta = theta
        self.omega = omega
        super().__init__("X", target, param_name="", noise=noise)

        self.register_buffer(
            "proj00",
            torch.tensor([[1, 0], [0, 0]], dtype=DEFAULT_MATRIX_DTYPE).unsqueeze(2),
        )
        self.register_buffer(
            "proj01",
            torch.tensor([[0, 1], [0, 0]], dtype=DEFAULT_MATRIX_DTYPE).unsqueeze(2),
        )
        self.register_buffer(
            "proj10",
            torch.tensor([[0, 0], [1, 0]], dtype=DEFAULT_MATRIX_DTYPE).unsqueeze(2),
        )
        self.register_buffer(
            "proj11",
            torch.tensor([[0, 0], [0, 1]], dtype=DEFAULT_MATRIX_DTYPE).unsqueeze(2),
        )

    @cached_property
    def eigenvals_generator(self) -> Tensor:
        """Get eigenvalues of the underlying generator.

        Arguments:
            values: Parameter values.

        Returns:
            Eigenvalues of the generator operator.
        """
        return pauli_singleq_eigenvalues.to(device=self.device, dtype=self.dtype)

    def _construct_parametric_base_op(
        self,
        values: dict[str, Tensor] | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        """
        Get the corresponding unitary.

        Arguments:
            values: Parameter value.
            embedding: An optional embedding for parameters.

        Returns:
            The unitary representation.
        """
        if embedding is not None:
            raise NotImplementedError()
        values = values or dict()
        phi, theta, omega = list(
            map(
                lambda t: t.unsqueeze(0) if len(t.size()) == 0 else t,
                [values[self.phi], values[self.theta], values[self.omega]],
            )
        )
        batch_size = len(theta)

        dtype = self.proj00.dtype
        device = self.proj00.device
        t_plus = torch.exp(-1j * (phi + omega) / 2).to(dtype=dtype)
        t_minus = torch.exp(-1j * (phi - omega) / 2).to(dtype=dtype)
        sin_t = (
            torch.sin(theta / 2)
            .unsqueeze(0)
            .unsqueeze(1)
            .repeat((2, 2, 1))
            .to(dtype=dtype, device=device)
        )
        cos_t = (
            torch.cos(theta / 2)
            .unsqueeze(0)
            .unsqueeze(1)
            .repeat((2, 2, 1))
            .to(dtype=dtype, device=device)
        )

        proj00 = self.proj00.repeat(1, 1, batch_size) * cos_t * t_plus
        proj01 = self.proj01.repeat(1, 1, batch_size) * sin_t * torch.conj(t_minus)
        proj10 = self.proj10.repeat(1, 1, batch_size) * sin_t * t_minus
        proj11 = self.proj11.repeat(1, 1, batch_size) * cos_t * torch.conj(t_plus)
        return proj00 - proj01 + proj10 + proj11

    def jacobian(
        self,
        values: dict[str, Tensor] | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        """
        Get the corresponding unitary of the jacobian.

        Arguments:
            values: Parameter value.

        Returns:
            The unitary representation of the jacobian.

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def digital_decomposition(self) -> list[Parametric]:
        """
        Get the digital decomposition of U.

        Returns:
            The digital decomposition.
        """
        target = self.target[0]
        return [
            RZ(target, self.phi),
            RY(target, self.theta),
            RZ(target, self.omega),
        ]

    def jacobian_decomposed(
        self, values: dict[str, Tensor] | None = None
    ) -> list[Tensor]:
        """
        Get the corresponding unitary decomposition of the jacobian.

        Arguments:
            values: Parameter value.

        Returns:
            The unitary decomposition of the jacobian.

        Raises:
            NotImplementedError
        """
        values = values or dict()
        return [op.jacobian(values) for op in self.digital_decomposition()]


OPS_PARAM_1Q = {PHASE, RX, RY, RZ, U}
OPS_PARAM_2Q = {CPHASE, CRX, CRY, CRZ}
OPS_PARAM = OPS_PARAM_1Q.union(OPS_PARAM_2Q)
OPS_DIAGONAL_PARAM = {RZ, PHASE, CRZ, CPHASE}



from __future__ import annotations

from typing import Callable

import numpy as np
import pytest
import qutip
import torch
from qutip import Qobj
from torch import Tensor

from pyqtorch.time_dependent.mesolve import mesolve
from pyqtorch.time_dependent.sesolve import sesolve
from pyqtorch.utils import SolverType

ATOL = 5e-2


@pytest.mark.flaky(max_runs=5)
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("ode_solver", [SolverType.DP5_SE, SolverType.KRYLOV_SE])
def test_sesolve(
    duration: float,
    batch_size: int,
    n_steps: int,
    torch_hamiltonian: Callable,
    qutip_hamiltonian: Callable,
    ode_solver: SolverType,
) -> None:
    psi0_qutip = qutip.basis(4, 0)

    # simulate with torch-based solver
    psi0_torch = (
        torch.tensor(psi0_qutip.full()).to(torch.complex128).repeat(1, batch_size)
    )

    t_points = torch.linspace(0, duration, n_steps)
    state_torch = sesolve(torch_hamiltonian, psi0_torch, t_points, ode_solver).states[
        -1
    ]

    # simulate with qutip solver
    t_points = np.linspace(0, duration, n_steps)
    result = qutip.sesolve(qutip_hamiltonian, psi0_qutip, t_points)
    state_qutip = torch.tensor(result.states[-1].full()).repeat(1, batch_size)

    assert torch.allclose(state_torch, state_qutip, atol=ATOL)


@pytest.mark.flaky(max_runs=5)
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("ode_solver", [SolverType.DP5_ME])
def test_mesolve(
    duration: float,
    batch_size: int,
    n_steps: int,
    torch_hamiltonian: Callable,
    qutip_hamiltonian: Callable,
    jump_op_torch: list[Tensor],
    jump_op_qutip: list[Qobj],
    ode_solver: SolverType,
) -> None:
    psi0_qutip = qutip.basis(4, 0)

    # simulate with torch-based solver
    psi0_torch = torch.tensor(psi0_qutip.full()).to(torch.complex128)
    rho0_torch = (
        torch.matmul(psi0_torch, psi0_torch.T).unsqueeze(-1).repeat(1, 1, batch_size)
    )

    t_points = torch.linspace(0, duration, n_steps)
    state_torch = mesolve(
        torch_hamiltonian, rho0_torch, jump_op_torch, t_points, ode_solver
    ).states[-1]

    # simulate with qutip solver
    t_points = np.linspace(0, duration, n_steps)
    result = qutip.mesolve(qutip_hamiltonian, psi0_qutip, t_points, jump_op_qutip)
    state_qutip = (
        torch.tensor(result.states[-1].full()).unsqueeze(-1).repeat(1, 1, batch_size)
    )

    assert torch.allclose(state_torch, state_qutip, atol=ATOL)



from __future__ import annotations

import warnings
from typing import Any, Callable

import torch
from torch import Tensor

from pyqtorch.time_dependent.integrators.adaptive import AdaptiveIntegrator
from pyqtorch.time_dependent.methods.dp5 import DormandPrince5
from pyqtorch.time_dependent.methods.krylov import Krylov
from pyqtorch.time_dependent.options import AdaptiveSolverOptions
from pyqtorch.utils import cache


class SESolver(AdaptiveIntegrator):
    """Generic adaptive-step solver for Schrodinger equation."""

    def __init__(
        self,
        H: Tensor | Callable[..., Any],
        psi0: Tensor,
        tsave: Tensor,
        options: AdaptiveSolverOptions,
    ):
        super().__init__(H, psi0, tsave, options)

    def ode_fun(self, t: float, psi: Tensor) -> Tensor:
        """Compute dpsi / dt = -1j * H(psi) at time t."""
        with warnings.catch_warnings():
            # filter-out UserWarning about "Sparse CSR tensor support is in beta state"
            warnings.filterwarnings("ignore", category=UserWarning)
            res = (
                -1j * self.H(t) @ (psi.to_sparse() if self.options.use_sparse else psi)
            )
        return res


class MESolver(AdaptiveIntegrator):
    """Generic adaptive-step solver for Linblad master equation."""

    def __init__(
        self,
        H: Callable[..., Any],
        rho0: Tensor,
        L: Tensor,
        tsave: Tensor,
        options: AdaptiveSolverOptions,
    ):
        super().__init__(H, rho0, tsave, options)

        self.L = L
        self.L_tuple = torch.unbind(self.L)

        L_concat_v = torch.cat(self.L_tuple)

        self.sum_LdagL = L_concat_v.mH @ L_concat_v

        # define identity operator
        n = self.H(0.0).size(-1)
        self.I = torch.eye(n, dtype=options.ctype)

        # define cached non-hermitian Hamiltonian
        self.Hnh = cache(lambda H: -0.5j * self.sum_LdagL + H)

    def ode_fun(self, t: float, rho: Tensor) -> Tensor:
        H = self.H(t)
        lindblad_term = sum(L @ rho @ L.mH for L in self.L_tuple)
        out: Tensor = -1j * self.Hnh(H) @ rho + 0.5 * lindblad_term
        return out + out.mH


class SEDormandPrince5(SESolver, DormandPrince5):
    pass


class MEDormandPrince5(MESolver, DormandPrince5):
    pass


class SEKrylov(Krylov):
    pass
